"""unidep - Unified Conda and Pip requirements management.

This module contains the Hatchling integration.
"""
from __future__ import annotations

from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.plugin import hookimpl

from unidep._setuptools_integration import get_python_dependencies
from unidep.utils import dependencies_filename, identify_current_platform

__all__ = ["UnidepRequirementsMetadataHook"]


class UnidepRequirementsMetadataHook(MetadataHookInterface):
    """Hatch hook to populate ``'project.depencencies'`` from ``requirements.yaml`` or ``pyproject.toml``."""  # noqa: E501

    PLUGIN_NAME = "unidep"

    def update(self, metadata: dict) -> None:
        """Update the project table's metadata."""
        if "dependencies" not in metadata.get("dynamic", []):
            return
        project_root = Path().resolve()
        try:
            requirements_file = dependencies_filename(project_root)
        except FileNotFoundError:
            return
        if "dependencies" in metadata:
            error_msg = (
                "You have a `requirements.yaml` file in your project root or"
                " configured unidep in `pyproject.toml` with `[tool.unidep]`,"
                " but you are also using `[project.dependencies]`."
                " Please remove `[project.dependencies]`, you cannot use both."
            )
            raise RuntimeError(error_msg)
        metadata["dependencies"] = get_python_dependencies(
            requirements_file,
            platforms=[identify_current_platform()],
            raises_if_missing=False,
        )


@hookimpl
def hatch_register_metadata_hook() -> type[UnidepRequirementsMetadataHook]:
    return UnidepRequirementsMetadataHook



"""Tests for the unidep.utils module."""
from __future__ import annotations

import sys
from unittest.mock import patch

import pytest

from unidep.platform_definitions import Selector
from unidep.utils import (
    UnsupportedPlatformError,
    build_pep508_environment_marker,
    escape_unicode,
    extract_matching_platforms,
    identify_current_platform,
    parse_package_str,
)

if sys.version_info >= (3, 8):
    from typing import get_args
else:  # pragma: no cover
    from typing_extensions import get_args


def test_escape_unicode() -> None:
    assert escape_unicode("foo\\n") == "foo\n"
    assert escape_unicode("foo\\t") == "foo\t"


def test_build_pep508_environment_marker() -> None:
    # Test with a single platform
    assert (
        build_pep508_environment_marker(["linux-64"])
        == "sys_platform == 'linux' and platform_machine == 'x86_64'"
    )

    # Test with multiple platforms
    assert (
        build_pep508_environment_marker(["linux-64", "osx-64"])
        == "sys_platform == 'linux' and platform_machine == 'x86_64' or sys_platform == 'darwin' and platform_machine == 'x86_64'"
    )

    # Test with an empty list
    assert not build_pep508_environment_marker([])

    # Test with a platform not in PEP508_MARKERS
    assert not build_pep508_environment_marker(["unknown-platform"])  # type: ignore[list-item]

    # Test with a mix of valid and invalid platforms
    assert (
        build_pep508_environment_marker(["linux-64", "unknown-platform"])  # type: ignore[list-item]
        == "sys_platform == 'linux' and platform_machine == 'x86_64'"
    )


def test_detect_platform() -> None:
    with patch("platform.system", return_value="Linux"), patch(
        "platform.machine",
        return_value="x86_64",
    ):
        assert identify_current_platform() == "linux-64"

    with patch("platform.system", return_value="Linux"), patch(
        "platform.machine",
        return_value="aarch64",
    ):
        assert identify_current_platform() == "linux-aarch64"

    with patch("platform.system", return_value="Darwin"), patch(
        "platform.machine",
        return_value="x86_64",
    ):
        assert identify_current_platform() == "osx-64"

    with patch("platform.system", return_value="Darwin"), patch(
        "platform.machine",
        return_value="arm64",
    ):
        assert identify_current_platform() == "osx-arm64"

    with patch("platform.system", return_value="Windows"), patch(
        "platform.machine",
        return_value="AMD64",
    ):
        assert identify_current_platform() == "win-64"

    with patch("platform.system", return_value="Linux"), patch(
        "platform.machine",
        return_value="unknown",
    ), pytest.raises(UnsupportedPlatformError, match="Unsupported Linux architecture"):
        identify_current_platform()

    with patch("platform.system", return_value="Darwin"), patch(
        "platform.machine",
        return_value="unknown",
    ), pytest.raises(UnsupportedPlatformError, match="Unsupported macOS architecture"):
        identify_current_platform()

    with patch("platform.system", return_value="Windows"), patch(
        "platform.machine",
        return_value="unknown",
    ), pytest.raises(
        UnsupportedPlatformError,
        match="Unsupported Windows architecture",
    ):
        identify_current_platform()

    with patch("platform.system", return_value="Linux"), patch(
        "platform.machine",
        return_value="ppc64le",
    ):
        assert identify_current_platform() == "linux-ppc64le"

    with patch("platform.system", return_value="Unknown"), patch(
        "platform.machine",
        return_value="x86_64",
    ), pytest.raises(UnsupportedPlatformError, match="Unsupported operating system"):
        identify_current_platform()


def test_parse_package_str() -> None:
    # Test with version pin
    assert parse_package_str("numpy >=1.20.0") == ("numpy", ">=1.20.0", None)
    assert parse_package_str("pandas<2.0,>=1.1.3") == ("pandas", "<2.0,>=1.1.3", None)

    # Test with multiple version conditions
    assert parse_package_str("scipy>=1.2.3, <1.3") == ("scipy", ">=1.2.3, <1.3", None)

    # Test with no version pin
    assert parse_package_str("matplotlib") == ("matplotlib", None, None)

    # Test with whitespace variations
    assert parse_package_str("requests >= 2.25") == ("requests", ">= 2.25", None)

    # Test when installing from a URL
    url = "https://github.com/python-adaptive/adaptive.git@main"
    pin = f"@ git+{url}"
    assert parse_package_str(f"adaptive {pin}") == ("adaptive", pin, None)

    # Test with invalid input
    with pytest.raises(ValueError, match="Invalid package string"):
        parse_package_str(">=1.20.0 numpy")


def test_parse_package_str_with_selector() -> None:
    # Test with version pin
    assert parse_package_str("numpy >=1.20.0:linux64") == (
        "numpy",
        ">=1.20.0",
        "linux64",
    )
    assert parse_package_str("pandas<2.0,>=1.1.3:osx") == (
        "pandas",
        "<2.0,>=1.1.3",
        "osx",
    )

    # Test with multiple version conditions
    assert parse_package_str("scipy>=1.2.3, <1.3:win") == (
        "scipy",
        ">=1.2.3, <1.3",
        "win",
    )

    # Test with no version pin
    assert parse_package_str("matplotlib:win") == ("matplotlib", None, "win")

    # Test with whitespace variations
    assert parse_package_str("requests >= 2.25:win") == ("requests", ">= 2.25", "win")

    # Test when installing from a URL
    url = "https://github.com/python-adaptive/adaptive.git@main"
    pin = f"@ git+{url}"
    assert parse_package_str(f"adaptive {pin}:win") == ("adaptive", pin, "win")

    for sel in get_args(Selector):
        assert parse_package_str(f"numpy:{sel}") == ("numpy", None, sel)

    # Test with multiple selectors
    assert parse_package_str("numpy:linux64 win64") == ("numpy", None, "linux64 win64")
    with pytest.raises(ValueError, match="Invalid platform selector: `unknown`"):
        assert parse_package_str("numpy:linux64 unknown")


def test_extract_matching_platforms() -> None:
    # Test with a line having a linux selector
    content_linux = "dependency1  # [linux]"
    assert set(extract_matching_platforms(content_linux)) == {
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
    }

    # Test with a line having a win selector
    content_win = "dependency2  # [win]"
    assert set(extract_matching_platforms(content_win)) == {"win-64"}

    # Test with a line having an osx64 selector
    content_osx64 = "dependency3  # [osx64]"
    assert set(extract_matching_platforms(content_osx64)) == {"osx-64"}

    # Test with a line having no selector
    content_none = "dependency4"
    assert extract_matching_platforms(content_none) == []

    # Test with a comment line
    content_comment = "# This is a comment"
    assert extract_matching_platforms(content_comment) == []

    # Test with a line having a unix selector
    content_unix = "dependency5  # [unix]"
    expected_unix = {
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
        "osx-64",
        "osx-arm64",
    }
    assert set(extract_matching_platforms(content_unix)) == expected_unix

    # Test with a line having multiple selectors
    content_multi = "dependency7  # [linux64 unix]"
    expected_multi = {
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
        "osx-64",
        "osx-arm64",
    }
    assert set(extract_matching_platforms(content_multi)) == expected_multi

    # Test with a line having multiple []
    content_multi = "dependency7  # [linux64] [win]"
    with pytest.raises(ValueError, match="Multiple bracketed selectors"):
        extract_matching_platforms(content_multi)

    incorrect_platform = "dependency8  # [unknown-platform]"
    with pytest.raises(ValueError, match="Invalid platform selector"):
        extract_matching_platforms(incorrect_platform)



"""unidep - Unified Conda and Pip requirements management."""

from unidep._conda_env import (
    create_conda_env_specification,
    write_conda_environment_file,
)
from unidep._conflicts import resolve_conflicts
from unidep._dependencies_parsing import (
    find_requirements_files,
    parse_local_dependencies,
    parse_requirements,
)
from unidep._setuptools_integration import (
    filter_python_dependencies,
    get_python_dependencies,
)
from unidep._version import __version__

__all__ = [
    "create_conda_env_specification",
    "filter_python_dependencies",
    "find_requirements_files",
    "get_python_dependencies",
    "parse_local_dependencies",
    "parse_requirements",
    "resolve_conflicts",
    "write_conda_environment_file",
    "__version__",
]



"""unidep - Unified Conda and Pip requirements management.

This module provides utility functions used throughout the package.
"""
from __future__ import annotations

import codecs
import platform
import re
import sys
import warnings
from pathlib import Path
from typing import NamedTuple, cast

from unidep._version import __version__
from unidep.platform_definitions import (
    PEP508_MARKERS,
    Platform,
    Selector,
    platforms_from_selector,
    validate_selector,
)

try:  # pragma: no cover
    if sys.version_info >= (3, 11):
        import tomllib
    else:
        import tomli as tomllib
    HAS_TOML = True
except ImportError:  # pragma: no cover
    HAS_TOML = False


def add_comment_to_file(
    filename: str | Path,
    extra_lines: list[str] | None = None,
) -> None:
    """Add a comment to the top of a file."""
    if extra_lines is None:
        extra_lines = []
    with open(filename, "r+") as f:  # noqa: PTH123
        content = f.read()
        f.seek(0, 0)
        command_line_args = " ".join(sys.argv[1:])
        txt = [
            f"# This file is created and managed by `unidep` {__version__}.",
            "# For details see https://github.com/basnijholt/unidep",
            f"# File generated with: `unidep {command_line_args}`",
            *extra_lines,
        ]
        content = "\n".join(txt) + "\n\n" + content
        f.write(content)


def remove_top_comments(filename: str | Path) -> None:
    """Removes the top comments (lines starting with '#') from a file."""
    with open(filename) as file:  # noqa: PTH123
        lines = file.readlines()

    first_non_comment = next(
        (i for i, line in enumerate(lines) if not line.strip().startswith("#")),
        len(lines),
    )
    content_without_comments = lines[first_non_comment:]
    with open(filename, "w") as file:  # noqa: PTH123
        file.writelines(content_without_comments)


def escape_unicode(string: str) -> str:
    """Escape unicode characters."""
    return codecs.decode(string, "unicode_escape")


def is_pip_installable(folder: str | Path) -> bool:  # pragma: no cover
    """Determine if the project is pip installable.

    Checks for existence of setup.py or [build-system] in pyproject.toml.
    """
    path = Path(folder)
    if (path / "setup.py").exists():
        return True

    # When toml makes it into the standard library, we can use that instead
    # For now this is good enough, except it doesn't handle the case where
    # [build-system] is inside of a multi-line literal string.
    pyproject_path = path / "pyproject.toml"
    if pyproject_path.exists():
        with pyproject_path.open("r") as file:
            for line in file:
                if line.strip().startswith("[build-system]"):
                    return True
    return False


class UnsupportedPlatformError(Exception):
    """Raised when the current platform is not supported."""


def identify_current_platform() -> Platform:
    """Detect the current platform."""
    system = platform.system().lower()
    architecture = platform.machine().lower()

    if system == "linux":
        if architecture == "x86_64":
            return "linux-64"
        if architecture == "aarch64":
            return "linux-aarch64"
        if architecture == "ppc64le":
            return "linux-ppc64le"
        msg = f"Unsupported Linux architecture `{architecture}`"
        raise UnsupportedPlatformError(msg)
    if system == "darwin":
        if architecture == "x86_64":
            return "osx-64"
        if architecture == "arm64":
            return "osx-arm64"
        msg = f"Unsupported macOS architecture `{architecture}`"
        raise UnsupportedPlatformError(msg)
    if system == "windows":
        if "64" in architecture:
            return "win-64"
        msg = f"Unsupported Windows architecture `{architecture}`"
        raise UnsupportedPlatformError(msg)
    msg = f"Unsupported operating system `{system}` with architecture `{architecture}`"
    raise UnsupportedPlatformError(msg)


def build_pep508_environment_marker(
    platforms: list[Platform | tuple[Platform, ...]],
) -> str:
    """Generate a PEP 508 selector for a list of platforms."""
    sorted_platforms = tuple(sorted(platforms))
    if sorted_platforms in PEP508_MARKERS:
        return PEP508_MARKERS[sorted_platforms]  # type: ignore[index]
    environment_markers = [
        PEP508_MARKERS[platform]
        for platform in sorted(sorted_platforms)
        if platform in PEP508_MARKERS
    ]
    return " or ".join(environment_markers)


class ParsedPackageStr(NamedTuple):
    """A package name and version pinning."""

    name: str
    pin: str | None = None
    # can be of type `Selector` but also space separated string of `Selector`s
    selector: str | None = None


def parse_package_str(package_str: str) -> ParsedPackageStr:
    """Splits a string into package name, version pinning, and platform selector."""
    # Regex to match package name, version pinning, and optionally platform selector
    name_pattern = r"[a-zA-Z0-9_.-]+"
    version_pin_pattern = r".*?"
    selector_pattern = r"[a-z0-9\s]+"
    pattern = rf"({name_pattern})\s*({version_pin_pattern})?(:({selector_pattern}))?$"
    match = re.match(pattern, package_str)

    if match:
        package_name = match.group(1).strip()
        version_pin = match.group(2).strip() if match.group(2) else None
        selector = match.group(4).strip() if match.group(4) else None

        if selector is not None:
            for s in selector.split():
                validate_selector(cast(Selector, s))

        return ParsedPackageStr(
            package_name,
            version_pin,
            selector,
        )

    msg = f"Invalid package string: '{package_str}'"
    raise ValueError(msg)


def _simple_warning_format(
    message: Warning | str,
    category: type[Warning],  # noqa: ARG001
    filename: str,
    lineno: int,
    line: str | None = None,  # noqa: ARG001
) -> str:  # pragma: no cover
    """Format warnings without code context."""
    return (
        f"---------------------\n"
        f"⚠️  *** WARNING *** ⚠️\n"
        f"{message}\n"
        f"Location: {filename}:{lineno}\n"
        f"---------------------\n"
    )


def warn(
    message: str | Warning,
    category: type[Warning] = UserWarning,
    stacklevel: int = 1,
) -> None:
    """Emit a warning with a custom format specific to this package."""
    original_format = warnings.formatwarning
    warnings.formatwarning = _simple_warning_format
    try:
        warnings.warn(message, category, stacklevel=stacklevel + 1)
    finally:
        warnings.formatwarning = original_format


def selector_from_comment(comment: str) -> str | None:
    """Extract a valid selector from a comment."""
    multiple_brackets_pat = re.compile(r"#.*\].*\[")  # Detects multiple brackets
    if multiple_brackets_pat.search(comment):
        msg = f"Multiple bracketed selectors found in comment: '{comment}'"
        raise ValueError(msg)

    sel_pat = re.compile(r"#\s*\[([^\[\]]+)\]")
    m = sel_pat.search(comment)
    if not m:
        return None
    selectors = m.group(1).strip().split()
    for s in selectors:
        validate_selector(cast(Selector, s))
    return " ".join(selectors)


def extract_matching_platforms(comment: str) -> list[Platform]:
    """Get all platforms matching a comment."""
    selector = selector_from_comment(comment)
    if selector is None:
        return []
    return platforms_from_selector(selector)


def unidep_configured_in_toml(path: Path) -> bool:
    """Check if dependencies are specified in pyproject.toml.

    If a TOML parser is not available it finds `[tool.unidep]` in `pyproject.toml`.
    """
    if HAS_TOML:
        with path.open("rb") as f:
            data = tomllib.load(f)
        return bool(data.get("tool", {}).get("unidep", {}))
    # TODO[Bas]: will fail if defining dict in  # noqa: TD004, TD003, FIX002
    # pyproject.toml directly e.g., it contains:
    # `tool = {unidep = {dependencies = ...}}`
    return any(  # pragma: no cover
        line.lstrip().startswith("[tool.unidep")
        for line in path.read_text().splitlines()
    )


def dependencies_filename(folder_or_path: str | Path) -> Path:
    """Get the path to `requirements.yaml` or `pyproject.toml` file."""
    path = Path(folder_or_path)
    if path.is_dir():
        fname_yaml = path / "requirements.yaml"
        if fname_yaml.exists():
            return fname_yaml
        fname_toml = path / "pyproject.toml"
        if fname_toml.exists() and unidep_configured_in_toml(fname_toml):
            return fname_toml
        msg = (
            f"File `{fname_yaml}` or `{fname_toml}` (with unidep configuration)"
            f" not found in `{folder_or_path}`."
        )
        raise FileNotFoundError(msg)
    if not path.exists():
        msg = f"File `{path}` not found."
        raise FileNotFoundError(msg)
    return path



"""unidep tests."""
from __future__ import annotations

import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from ruamel.yaml import YAML

from unidep import (
    create_conda_env_specification,
    filter_python_dependencies,
    find_requirements_files,
    get_python_dependencies,
    parse_requirements,
    resolve_conflicts,
    write_conda_environment_file,
)
from unidep._conda_env import CondaEnvironmentSpec
from unidep._conflicts import VersionConflictError
from unidep._dependencies_parsing import yaml_to_toml
from unidep.platform_definitions import Platform, Spec

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:  # pragma: no cover
        from typing_extensions import Literal


REPO_ROOT = Path(__file__).parent.parent


def maybe_as_toml(toml_or_yaml: Literal["toml", "yaml"], p: Path) -> Path:
    if toml_or_yaml == "toml":
        toml = yaml_to_toml(p)
        p.unlink()
        p = p.with_name("pyproject.toml")
        p.write_text(toml)
    return p


@pytest.fixture(params=["toml", "yaml"])
def setup_test_files(
    request: pytest.FixtureRequest,
    tmp_path: Path,
) -> tuple[Path, Path]:
    d1 = tmp_path / "dir1"
    d1.mkdir()
    f1 = d1 / "requirements.yaml"
    f1.write_text("dependencies:\n  - numpy\n  - conda: mumps")

    d2 = tmp_path / "dir2"
    d2.mkdir()
    f2 = d2 / "requirements.yaml"
    f2.write_text("dependencies:\n  - pip: pandas")
    f1 = maybe_as_toml(request.param, f1)
    f2 = maybe_as_toml(request.param, f2)
    return (f1, f2)


def test_find_requirements_files(
    tmp_path: Path,
    setup_test_files: tuple[Path, Path],
) -> None:
    # Make sure to pass the depth argument correctly if your function expects it.
    found_files = find_requirements_files(
        tmp_path,
        depth=1,
        verbose=True,
    )

    # Convert found_files to absolute paths for comparison
    absolute_results = sorted(str(p.resolve()) for p in found_files)
    absolute_test_files = sorted(str(p.resolve()) for p in setup_test_files)

    assert absolute_results == absolute_test_files


def test_find_requirements_files_depth(tmp_path: Path) -> None:
    # Create a nested directory structure
    (tmp_path / "dir1").mkdir()
    (tmp_path / "dir1/dir2").mkdir()
    (tmp_path / "dir1/dir2/dir3").mkdir()

    # Create test files
    (tmp_path / "requirements.yaml").touch()
    (tmp_path / "dir1/requirements.yaml").touch()
    (tmp_path / "dir1/dir2/requirements.yaml").touch()
    (tmp_path / "dir1/dir2/dir3/requirements.yaml").touch()

    # Test depth=0
    assert len(find_requirements_files(tmp_path, depth=0)) == 1

    # Test depth=1
    assert len(find_requirements_files(tmp_path, depth=1)) == 2

    # Test depth=2
    assert len(find_requirements_files(tmp_path, depth=2)) == 3

    # Test depth=3
    assert len(find_requirements_files(tmp_path, depth=3)) == 4

    # Test depth=4 (or more)
    assert len(find_requirements_files(tmp_path, depth=4)) == 4


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_requirements(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1 # [linux64]
                - foo # [unix]
                - bar >1
                - bar
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="conda",
                selector="unix",
                identifier="530d9eaa",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="unix",
                identifier="530d9eaa",
            ),
        ],
        "bar": [
            Spec(
                name="bar",
                which="conda",
                pin=">1",
                identifier="08fd8713",
            ),
            Spec(
                name="bar",
                which="pip",
                pin=">1",
                identifier="08fd8713",
            ),
            Spec(
                name="bar",
                which="conda",
                identifier="9e467fa1",
            ),
            Spec(
                name="bar",
                which="pip",
                identifier="9e467fa1",
            ),
        ],
    }


@pytest.mark.parametrize("verbose", [True, False])
def test_generate_conda_env_file(
    tmp_path: Path,
    verbose: bool,  # noqa: FBT001
    setup_test_files: tuple[Path, Path],
) -> None:
    output_file = tmp_path / "environment.yaml"
    requirements = parse_requirements(*setup_test_files, verbose=verbose)
    resolved = resolve_conflicts(
        requirements.requirements,
        requirements.platforms,
    )
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )

    write_conda_environment_file(env_spec, str(output_file), verbose=verbose)

    with output_file.open() as f, YAML(typ="safe") as yaml:
        env_data = yaml.load(f)
        assert "dependencies" in env_data
        assert "numpy" in env_data["dependencies"]
        assert {"pip": ["pandas"]} in env_data["dependencies"]


def test_generate_conda_env_stdout(
    setup_test_files: tuple[Path, Path],
    capsys: pytest.CaptureFixture,
) -> None:
    requirements = parse_requirements(*setup_test_files)
    resolved = resolve_conflicts(
        requirements.requirements,
        requirements.platforms,
    )
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    write_conda_environment_file(env_spec, output_file=None)
    captured = capsys.readouterr()
    assert "dependencies" in captured.out
    assert "numpy" in captured.out
    assert "- pandas" in captured.out


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_create_conda_env_specification_platforms(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - yolo  # [arm64]
                - foo  # [linux64]
                - conda: bar  # [win]
                - pip: pip-package
                - pip: pip-package2  # [arm64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p)
    resolved = resolve_conflicts(
        requirements.requirements,
        requirements.platforms,
    )
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == [
        {"sel(osx)": "yolo"},
        {"sel(linux)": "foo"},
        {"sel(win)": "bar"},
    ]
    expected_pip = [
        "pip-package",
        "pip-package2; sys_platform == 'darwin' and platform_machine == 'arm64'",
    ]
    assert env_spec.pip == expected_pip

    # Test on two platforms
    platforms: list[Platform] = ["osx-arm64", "win-64"]
    resolved = resolve_conflicts(
        requirements.requirements,
        platforms,
    )

    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms,
    )
    assert env_spec.conda == [{"sel(osx)": "yolo"}, {"sel(win)": "bar"}]
    assert env_spec.pip == expected_pip

    # Test with comment selector
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms,
        selector="comment",
    )
    assert env_spec.conda == ["yolo", "bar"]
    assert env_spec.pip == ["pip-package", "pip-package2"]
    write_conda_environment_file(env_spec, str(tmp_path / "environment.yaml"))
    with (tmp_path / "environment.yaml").open() as f:
        text = "".join(f.readlines())
        assert "- yolo  # [arm64]" in text
        assert "- bar # [win64]" in text

    with pytest.raises(ValueError, match="Invalid platform"):
        resolve_conflicts(
            requirements.requirements,
            ["unknown-platform"],  # type: ignore[list-item]
        )


def test_verbose_output(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    f = tmp_path / "dir3" / "requirements.yaml"
    f.parent.mkdir()
    f.write_text("dependencies:\n  - scipy")

    find_requirements_files(tmp_path, verbose=True)
    captured = capsys.readouterr()
    assert "Scanning in" in captured.out
    assert str(tmp_path / "dir3") in captured.out

    parse_requirements(f, verbose=True)
    captured = capsys.readouterr()
    assert "Parsing" in captured.out
    assert str(f) in captured.out

    write_conda_environment_file(
        CondaEnvironmentSpec(channels=[], platforms=[], conda=[], pip=[]),
        verbose=True,
    )
    captured = capsys.readouterr()
    assert "Generating environment file at" in captured.out
    assert "Environment file generated successfully." in captured.out


def test_extract_python_requires(setup_test_files: tuple[Path, Path]) -> None:
    f1, f2 = setup_test_files
    requires1 = get_python_dependencies(str(f1))
    assert requires1 == ["numpy"]
    requires2 = get_python_dependencies(str(f2))
    assert requires2 == ["pandas"]

    # Test with a file that doesn't exist
    with pytest.raises(FileNotFoundError):
        get_python_dependencies("nonexistent_file.yaml", raises_if_missing=True)
    assert (
        get_python_dependencies("nonexistent_file.yaml", raises_if_missing=False) == []
    )


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_channels(toml_or_yaml: Literal["toml", "yaml"], tmp_path: Path) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text("channels:\n  - conda-forge\n  - defaults")
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.channels == ["conda-forge", "defaults"]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_surrounding_comments(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
            # This is a comment before
                - yolo  # [osx]
            # This is a comment after
                # This is another comment
                - foo  # [linux]
                # And this is a comment after
                - bar  # [win]
                # Next is an empty comment
                - baz  #
                - pip: pip-package
                #
                - pip: pip-package2  # [osx]
                #
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "yolo": [
            Spec(
                name="yolo",
                which="conda",
                selector="osx",
                identifier="8b0c4c31",
            ),
            Spec(
                name="yolo",
                which="pip",
                selector="osx",
                identifier="8b0c4c31",
            ),
        ],
        "foo": [
            Spec(
                name="foo",
                which="conda",
                selector="linux",
                identifier="ecd4baa6",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux",
                identifier="ecd4baa6",
            ),
        ],
        "bar": [
            Spec(
                name="bar",
                which="conda",
                selector="win",
                identifier="8528de75",
            ),
            Spec(
                name="bar",
                which="pip",
                selector="win",
                identifier="8528de75",
            ),
        ],
        "baz": [
            Spec(
                name="baz",
                which="conda",
                identifier="9e467fa1",
            ),
            Spec(name="baz", which="pip", identifier="9e467fa1"),
        ],
        "pip-package": [
            Spec(
                name="pip-package",
                which="pip",
                identifier="5813b64a",
            ),
        ],
        "pip-package2": [
            Spec(
                name="pip-package2",
                which="pip",
                selector="osx",
                identifier="1c0fa4c4",
            ),
        ],
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_filter_pip_and_conda(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    # Setup a sample ParsedRequirements instance with platform selectors
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
              - conda: package1  # [linux64]
              - conda: package2  # [osx64]
              - pip: package3
              - pip: package4  # [unix]
              - common_package  # [unix]
              - conda: shared_package  # [linux64]
                pip: shared_package  # [win64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "package1": [
            Spec(
                name="package1",
                which="conda",
                selector="linux64",
                identifier="c292b98a",
            ),
        ],
        "package2": [
            Spec(
                name="package2",
                which="conda",
                selector="osx64",
                identifier="b2ac468f",
            ),
        ],
        "package3": [
            Spec(
                name="package3",
                which="pip",
                identifier="08fd8713",
            ),
        ],
        "package4": [
            Spec(
                name="package4",
                which="pip",
                selector="unix",
                identifier="1d5d7757",
            ),
        ],
        "common_package": [
            Spec(
                name="common_package",
                which="conda",
                selector="unix",
                identifier="f78244dc",
            ),
            Spec(
                name="common_package",
                which="pip",
                selector="unix",
                identifier="f78244dc",
            ),
        ],
        "shared_package": [
            Spec(
                name="shared_package",
                which="conda",
                selector="linux64",
                identifier="1599d575",
            ),
            Spec(
                name="shared_package",
                which="pip",
                selector="win64",
                identifier="46630b59",
            ),
        ],
    }

    resolved = resolve_conflicts(
        requirements.requirements,
        requirements.platforms,
    )
    assert resolved == {
        "package1": {
            "linux-64": {
                "conda": Spec(
                    name="package1",
                    which="conda",
                    selector="linux64",
                    identifier="c292b98a",
                ),
            },
        },
        "package2": {
            "osx-64": {
                "conda": Spec(
                    name="package2",
                    which="conda",
                    selector="osx64",
                    identifier="b2ac468f",
                ),
            },
        },
        "package3": {
            None: {
                "pip": Spec(
                    name="package3",
                    which="pip",
                    identifier="08fd8713",
                ),
            },
        },
        "package4": {
            "linux-64": {
                "pip": Spec(
                    name="package4",
                    which="pip",
                    selector="unix",
                    identifier="1d5d7757",
                ),
            },
            "linux-aarch64": {
                "pip": Spec(
                    name="package4",
                    which="pip",
                    selector="unix",
                    identifier="1d5d7757",
                ),
            },
            "linux-ppc64le": {
                "pip": Spec(
                    name="package4",
                    which="pip",
                    selector="unix",
                    identifier="1d5d7757",
                ),
            },
            "osx-64": {
                "pip": Spec(
                    name="package4",
                    which="pip",
                    selector="unix",
                    identifier="1d5d7757",
                ),
            },
            "osx-arm64": {
                "pip": Spec(
                    name="package4",
                    which="pip",
                    selector="unix",
                    identifier="1d5d7757",
                ),
            },
        },
        "common_package": {
            "linux-64": {
                "conda": Spec(
                    name="common_package",
                    which="conda",
                    selector="unix",
                    identifier="f78244dc",
                ),
                "pip": Spec(
                    name="common_package",
                    which="pip",
                    selector="unix",
                    identifier="f78244dc",
                ),
            },
            "linux-aarch64": {
                "conda": Spec(
                    name="common_package",
                    which="conda",
                    selector="unix",
                    identifier="f78244dc",
                ),
                "pip": Spec(
                    name="common_package",
                    which="pip",
                    selector="unix",
                    identifier="f78244dc",
                ),
            },
            "linux-ppc64le": {
                "conda": Spec(
                    name="common_package",
                    which="conda",
                    selector="unix",
                    identifier="f78244dc",
                ),
                "pip": Spec(
                    name="common_package",
                    which="pip",
                    selector="unix",
                    identifier="f78244dc",
                ),
            },
            "osx-64": {
                "conda": Spec(
                    name="common_package",
                    which="conda",
                    selector="unix",
                    identifier="f78244dc",
                ),
                "pip": Spec(
                    name="common_package",
                    which="pip",
                    selector="unix",
                    identifier="f78244dc",
                ),
            },
            "osx-arm64": {
                "conda": Spec(
                    name="common_package",
                    which="conda",
                    selector="unix",
                    identifier="f78244dc",
                ),
                "pip": Spec(
                    name="common_package",
                    which="pip",
                    selector="unix",
                    identifier="f78244dc",
                ),
            },
        },
        "shared_package": {
            "linux-64": {
                "conda": Spec(
                    name="shared_package",
                    which="conda",
                    selector="linux64",
                    identifier="1599d575",
                ),
            },
            "win-64": {
                "pip": Spec(
                    name="shared_package",
                    which="pip",
                    selector="win64",
                    identifier="46630b59",
                ),
            },
        },
    }
    # Pip
    python_deps = filter_python_dependencies(resolved)
    assert python_deps == [
        "common_package; sys_platform == 'linux' or sys_platform == 'darwin'",
        "package3",
        "package4; sys_platform == 'linux' or sys_platform == 'darwin'",
        "shared_package; sys_platform == 'win32' and platform_machine == 'AMD64'",
    ]

    # Conda
    conda_env_spec = create_conda_env_specification(
        resolved,
        channels=requirements.channels,
        platforms=requirements.platforms,
    )

    def sort(x: list[dict[str, str]]) -> list[dict[str, str]]:
        return sorted(x, key=lambda x: tuple(x.items()))

    assert sort(conda_env_spec.conda) == sort(  # type: ignore[arg-type]
        [
            {"sel(linux)": "package1"},
            {"sel(osx)": "package2"},
            {"sel(osx)": "common_package"},
            {"sel(linux)": "common_package"},
            {"sel(linux)": "shared_package"},
        ],
    )
    assert conda_env_spec.pip == [
        "package3",
        "package4; sys_platform == 'linux' or sys_platform == 'darwin'",
        "shared_package; sys_platform == 'win32' and platform_machine == 'AMD64'",
    ]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_duplicates_with_version(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1 # [linux64]
                - foo # [linux64]
                - bar
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="conda",
                selector="linux64",
                identifier="dd6a8aaf",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux64",
                identifier="dd6a8aaf",
            ),
        ],
        "bar": [
            Spec(
                name="bar",
                which="conda",
                identifier="08fd8713",
            ),
            Spec(
                name="bar",
                which="pip",
                identifier="08fd8713",
            ),
        ],
    }
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "foo": {
            "linux-64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    selector="linux64",
                    pin=">1",
                    identifier="c292b98a",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    selector="linux64",
                    pin=">1",
                    identifier="c292b98a",
                ),
            },
        },
        "bar": {
            None: {
                "conda": Spec(
                    name="bar",
                    which="conda",
                    identifier="08fd8713",
                ),
                "pip": Spec(
                    name="bar",
                    which="pip",
                    identifier="08fd8713",
                ),
            },
        },
    }
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == [{"sel(linux)": "foo >1"}, "bar"]
    assert env_spec.pip == []

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == [
        "bar",
        "foo >1; sys_platform == 'linux' and platform_machine == 'x86_64'",
    ]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_duplicates_different_platforms(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1 # [linux64]
                - foo <=2 # [linux]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="conda",
                selector="linux",
                pin="<=2",
                identifier="ecd4baa6",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux",
                pin="<=2",
                identifier="ecd4baa6",
            ),
        ],
    }
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "foo": {
            "linux-64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin=">1,<=2",
                    identifier="c292b98a",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin=">1,<=2",
                    identifier="c292b98a",
                ),
            },
            "linux-aarch64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    selector="linux",
                    pin="<=2",
                    identifier="ecd4baa6",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    selector="linux",
                    pin="<=2",
                    identifier="ecd4baa6",
                ),
            },
            "linux-ppc64le": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    selector="linux",
                    pin="<=2",
                    identifier="ecd4baa6",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    selector="linux",
                    pin="<=2",
                    identifier="ecd4baa6",
                ),
            },
        },
    }
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == [{"sel(linux)": "foo >1,<=2"}]
    assert env_spec.pip == []

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == [
        "foo <=2; sys_platform == 'linux' and platform_machine == 'aarch64'",
        "foo <=2; sys_platform == 'linux' and platform_machine == 'ppc64le'",
        "foo >1,<=2; sys_platform == 'linux' and platform_machine == 'x86_64'",
    ]

    # now only use linux-64
    platforms: list[Platform] = ["linux-64"]
    resolved = resolve_conflicts(requirements.requirements, platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms,
    )
    assert env_spec.conda == ["foo >1,<=2"]
    assert env_spec.pip == []


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_expand_none_with_different_platforms(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1 # [linux64]
                - foo <3
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="pip",
                selector="linux64",
                pin=">1",
                identifier="c292b98a",
            ),
            Spec(
                name="foo",
                which="conda",
                pin="<3",
                identifier="5eb93b8c",
            ),
            Spec(
                name="foo",
                which="pip",
                pin="<3",
                identifier="5eb93b8c",
            ),
        ],
    }
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "foo": {
            "linux-64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin=">1,<3",
                    identifier="c292b98a",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin=">1,<3",
                    identifier="c292b98a",
                ),
            },
            "linux-aarch64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
            },
            "linux-ppc64le": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
            },
            "osx-64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
            },
            "osx-arm64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
            },
            "win-64": {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin="<3",
                    identifier="5eb93b8c",
                ),
            },
        },
    }
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == [
        {"sel(linux)": "foo >1,<3"},
        {"sel(osx)": "foo <3"},
        {"sel(win)": "foo <3"},
    ]

    assert env_spec.pip == []

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == [
        "foo <3; sys_platform == 'darwin' and platform_machine == 'arm64'",
        "foo <3; sys_platform == 'darwin' and platform_machine == 'x86_64'",
        "foo <3; sys_platform == 'linux' and platform_machine == 'aarch64'",
        "foo <3; sys_platform == 'linux' and platform_machine == 'ppc64le'",
        "foo <3; sys_platform == 'win32' and platform_machine == 'AMD64'",
        "foo >1,<3; sys_platform == 'linux' and platform_machine == 'x86_64'",
    ]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_different_pins_on_conda_and_pip(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: foo >1
                  conda: foo <1
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                pin="<1",
                identifier="17e5d607",
            ),
            Spec(
                name="foo",
                which="pip",
                pin=">1",
                identifier="17e5d607",
            ),
        ],
    }
    with pytest.warns(UserWarning, match="Version Pinning Conflict"):
        resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "foo": {
            None: {
                "conda": Spec(
                    name="foo",
                    which="conda",
                    pin="<1",
                    identifier="17e5d607",
                ),
                "pip": Spec(
                    name="foo",
                    which="pip",
                    pin=">1",
                    identifier="17e5d607",
                ),
            },
        },
    }
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == ["foo <1"]

    assert env_spec.pip == []

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == ["foo >1"]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_pip_pinned_conda_not(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: foo >1
                  conda: foo
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == []

    assert env_spec.pip == ["foo >1"]

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == ["foo >1"]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_conda_pinned_pip_not(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: foo
                  conda: foo >1
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == ["foo >1"]

    assert env_spec.pip == []

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == []


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_filter_python_dependencies_with_platforms(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo # [unix]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, ["linux-64"])
    python_deps = filter_python_dependencies(resolved)
    assert python_deps == [
        "foo; sys_platform == 'linux' and platform_machine == 'x86_64'",
    ]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_conda_with_comments(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - adaptive # [linux64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
        selector="comment",
    )
    assert env_spec.conda == ["adaptive"]
    assert env_spec.pip == []
    write_conda_environment_file(env_spec, str(tmp_path / "environment.yaml"))
    with (tmp_path / "environment.yaml").open() as f:
        lines = f.readlines()
        dependency_line = next(line for line in lines if "adaptive" in line)
        assert "- adaptive  # [linux64]" in dependency_line


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_duplicate_names(toml_or_yaml: Literal["toml", "yaml"], tmp_path: Path) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - conda: flatbuffers
                - pip: flatbuffers
                  conda: python-flatbuffers
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == ["flatbuffers", "python-flatbuffers"]
    assert env_spec.pip == []

    python_deps = filter_python_dependencies(resolved)
    assert python_deps == ["flatbuffers"]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_conflicts_when_selector_comment(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1 # [linux64]
                - foo <2 # [linux]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
        selector="comment",
    )
    assert env_spec.conda == ["foo >1,<2", "foo <2", "foo <2"]
    assert env_spec.pip == []

    write_conda_environment_file(env_spec, str(tmp_path / "environment.yaml"))

    with (tmp_path / "environment.yaml").open() as f:
        text = "".join(f.readlines())
        assert "- foo >1,<2  # [linux64]" in text
        assert "- foo <2 # [aarch64]" in text
        assert "- foo <2 # [ppc64le]" in text

    # With just [unix]
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1
                - foo <2 # [unix]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
        selector="comment",
    )
    assert env_spec.conda == [
        "foo <2,>1",
        "foo <2,>1",
        "foo <2,>1",
        "foo <2,>1",
        "foo <2,>1",
        "foo >1",
    ]
    assert env_spec.pip == []

    write_conda_environment_file(env_spec, str(tmp_path / "environment.yaml"))

    with (tmp_path / "environment.yaml").open() as f:
        text = "".join(f.readlines())
        assert "- foo <2,>1  # [linux64]" in text
        assert "- foo <2,>1 # [osx64]" in text
        assert "- foo <2,>1 # [arm64]" in text
        assert "- foo <2,>1 # [aarch64]" in text
        assert "- foo <2,>1 # [ppc64le]" in text
        assert "- foo >1 # [win64]" in text


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_platforms_section_in_yaml(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            platforms:
                - linux-64
                - osx-arm64
            dependencies:
                - foo
                - bar # [win]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
        selector="sel",
    )
    assert env_spec.conda == ["foo"]
    assert env_spec.pip == []
    assert env_spec.platforms == ["linux-64", "osx-arm64"]
    python_deps = filter_python_dependencies(resolved)
    assert python_deps == ["foo"]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_platforms_section_in_yaml_similar_platforms(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            channels:
                - conda-forge
            platforms:
                - linux-64
                - linux-aarch64
            dependencies:
                - foo
                - bar # [win]
                - yolo <1 # [aarch64]
                - yolo >1 # [linux64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    with pytest.warns(UserWarning, match="Dependency Conflict on"):
        env_spec = create_conda_env_specification(
            resolved,
            requirements.channels,
            requirements.platforms,
            selector="sel",
        )
    assert env_spec.conda == ["foo", {"sel(linux)": "yolo <1"}]
    assert env_spec.pip == []
    assert env_spec.platforms == ["linux-64", "linux-aarch64"]
    python_deps = filter_python_dependencies(resolved)
    assert python_deps == [
        "foo",
        "yolo <1; sys_platform == 'linux' and platform_machine == 'aarch64'",
        "yolo >1; sys_platform == 'linux' and platform_machine == 'x86_64'",
    ]

    # Test with comment selector
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
        selector="comment",
    )
    assert env_spec.conda == ["foo", "yolo >1", "yolo <1"]
    assert env_spec.pip == []

    write_conda_environment_file(env_spec, str(tmp_path / "environment.yaml"))

    with (tmp_path / "environment.yaml").open() as f:
        text = "".join(f.readlines())
        assert "- yolo >1  # [linux64]" in text
        assert "- yolo <1 # [aarch64]" in text
        assert "platforms:" in text
        assert "- linux-64" in text
        assert "- linux-aarch64" in text


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_conda_with_non_platform_comment(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            channels:
                - conda-forge
            dependencies:
                - pip: qsimcirq  # [linux64]
                - pip: slurm-usage  # added to avoid https://github.com/conda/conda-lock/pull/564
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
        selector="comment",
    )
    assert env_spec.conda == []
    assert env_spec.pip == ["qsimcirq", "slurm-usage"]
    write_conda_environment_file(env_spec, str(tmp_path / "environment.yaml"))
    with (tmp_path / "environment.yaml").open() as f:
        lines = "".join(f.readlines())
    assert "- qsimcirq  # [linux64]" in lines
    assert "- slurm-usage" in lines
    assert "  - pip:" in lines


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_pip_and_conda_different_name_on_linux64(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    # On linux64, the conda package is called "cuquantum-python" and
    # the pip package is called "cuquantum". We test that not both
    # packages are in the final environment file.
    p.write_text(
        textwrap.dedent(
            """\
            name: test
            channels:
              - conda-forge
            dependencies:
              - conda: cuquantum-python  # [linux64]
                pip: cuquantum  # [linux64]
            platforms:
              - linux-64
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=True)
    expected = {
        "cuquantum-python": [
            Spec(
                name="cuquantum-python",
                which="conda",
                selector="linux64",
                identifier="c292b98a",
            ),
        ],
        "cuquantum": [
            Spec(
                name="cuquantum",
                which="pip",
                selector="linux64",
                identifier="c292b98a",
            ),
        ],
    }
    assert requirements.requirements == expected
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    expected_resolved = {
        "cuquantum-python": {
            "linux-64": {
                "conda": Spec(
                    name="cuquantum-python",
                    which="conda",
                    selector="linux64",
                    identifier="c292b98a",
                ),
            },
        },
        "cuquantum": {
            "linux-64": {
                "pip": Spec(
                    name="cuquantum",
                    which="pip",
                    selector="linux64",
                    identifier="c292b98a",
                ),
            },
        },
    }
    assert resolved == expected_resolved
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == ["cuquantum-python"]
    assert env_spec.pip == []


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_requirements_with_ignore_pin(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, ignore_pins=["foo"], verbose=False)
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                identifier="17e5d607",
            ),
            Spec(
                name="foo",
                which="pip",
                identifier="17e5d607",
            ),
        ],
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_requirements_with_skip_dependency(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1
                - bar
                - baz
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(
        p,
        skip_dependencies=["foo", "bar"],
        verbose=False,
    )
    assert requirements.requirements == {
        "baz": [
            Spec(
                name="baz",
                which="conda",
                identifier="08fd8713",
            ),
            Spec(
                name="baz",
                which="pip",
                identifier="08fd8713",
            ),
        ],
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_pin_star_cuda(toml_or_yaml: Literal["toml", "yaml"], tmp_path: Path) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - conda: qsimcirq * cuda*  # [linux64]
                - conda: qsimcirq * cpu*  # [arm64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p)
    assert requirements.requirements == {
        "qsimcirq": [
            Spec(
                name="qsimcirq",
                which="conda",
                selector="linux64",
                pin="* cuda*",
                identifier="c292b98a",
            ),
            Spec(
                name="qsimcirq",
                which="conda",
                selector="arm64",
                pin="* cpu*",
                identifier="489f33e0",
            ),
        ],
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_requirements_with_overwrite_pins(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo >1
                - conda: bar * cuda*
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(
        p,
        overwrite_pins=["foo=1", "bar * cpu*"],
        verbose=False,
    )
    assert requirements.requirements == {
        "foo": [
            Spec(
                name="foo",
                which="conda",
                pin="=1",
                identifier="17e5d607",
            ),
            Spec(
                name="foo",
                which="pip",
                pin="=1",
                identifier="17e5d607",
            ),
        ],
        "bar": [
            Spec(
                name="bar",
                which="conda",
                pin="* cpu*",
                identifier="5eb93b8c",
            ),
        ],
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_duplicate_names_different_platforms(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: ray  # [arm64]
                - conda: ray-core  # [linux64]
                  pip: ray # [linux64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(
        p,
        overwrite_pins=["foo=1", "bar * cpu*"],
        verbose=False,
    )
    assert requirements.requirements == {
        "ray": [
            Spec(
                name="ray",
                which="pip",
                selector="arm64",
                identifier="1b26c5b2",
            ),
            Spec(
                name="ray",
                which="pip",
                selector="linux64",
                identifier="dd6a8aaf",
            ),
        ],
        "ray-core": [
            Spec(
                name="ray-core",
                which="conda",
                selector="linux64",
                identifier="dd6a8aaf",
            ),
        ],
    }
    platforms_arm64: list[Platform] = ["osx-arm64"]
    resolved = resolve_conflicts(requirements.requirements, platforms_arm64)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms_arm64,
    )
    assert env_spec.conda == []
    assert env_spec.pip == ["ray"]

    platforms_linux64: list[Platform] = ["linux-64"]
    resolved = resolve_conflicts(requirements.requirements, platforms_linux64)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms_linux64,
    )
    assert env_spec.conda == ["ray-core"]
    assert env_spec.pip == []


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_with_unused_platform(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - adaptive # [linux64]
                - rsync-time-machine >0.1 # [osx64]
                - rsync-time-machine <3
                - rsync-time-machine >1 # [linux64]
            """,
        ),
    )
    p = maybe_as_toml(toml_or_yaml, p)
    requirements = parse_requirements(p, verbose=False)
    platforms: list[Platform] = ["linux-64"]
    resolved = resolve_conflicts(requirements.requirements, platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms,
        selector="comment",
    )
    assert env_spec.conda == ["adaptive", "rsync-time-machine >1,<3"]
    assert env_spec.pip == []


def test_pip_with_pinning(tmp_path: Path) -> None:
    p1 = tmp_path / "p1" / "requirements.yaml"
    p1.parent.mkdir()
    p1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: qiskit-terra ==0.25.2.1
                - pip: qiskit-terra ==0.25.2.2
            """,
        ),
    )

    requirements = parse_requirements(p1, verbose=False)
    with pytest.raises(
        VersionConflictError,
        match="Invalid version pinning '==0.25.2.1' for 'qiskit-terra'",
    ):
        resolve_conflicts(requirements.requirements, requirements.platforms)

    p2 = tmp_path / "p2" / "requirements.yaml"
    p2.parent.mkdir()
    p2.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: qiskit-terra =0.25.2.1
                - pip: qiskit-terra =0.25.2.1
            """,
        ),
    )

    requirements = parse_requirements(p2, verbose=False)
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        requirements.platforms,
    )
    assert env_spec.conda == []
    assert env_spec.pip == ["qiskit-terra ==0.25.2.1"]


def test_pip_with_pinning_special_case_wildcard(tmp_path: Path) -> None:
    p1 = tmp_path / "p1" / "requirements.yaml"
    p1.parent.mkdir()
    p1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: qsimcirq * cuda*
                - pip: qsimcirq * cuda*
            """,
        ),
    )

    requirements = parse_requirements(p1, verbose=False)

    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "qsimcirq": {
            None: {
                "pip": Spec(
                    name="qsimcirq",
                    which="pip",
                    pin="* cuda*",
                    identifier="17e5d607",
                ),
            },
        },
    }

    p2 = tmp_path / "p2" / "requirements.yaml"
    p2.parent.mkdir()
    p2.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: qsimcirq * cuda*
                - pip: qsimcirq * cpu*
            """,
        ),
    )

    requirements = parse_requirements(p2, verbose=False)

    with pytest.raises(
        VersionConflictError,
        match="['* cuda*', '* cpu*']",
    ):
        resolve_conflicts(requirements.requirements, requirements.platforms)


def test_pip_with_pinning_special_case_git_repo(tmp_path: Path) -> None:
    p1 = tmp_path / "p1" / "requirements.yaml"
    p1.parent.mkdir()
    p1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pip: adaptive @ git+https://github.com/python-adaptive/adaptive.git@main
                - pip: adaptive @ git+https://github.com/python-adaptive/adaptive.git@main
            """,
        ),
    )

    requirements = parse_requirements(p1, verbose=False)

    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "adaptive": {
            None: {
                "pip": Spec(
                    name="adaptive",
                    which="pip",
                    pin="@ git+https://github.com/python-adaptive/adaptive.git@main",
                    identifier="17e5d607",
                ),
            },
        },
    }


def test_not_equal(tmp_path: Path) -> None:
    p1 = tmp_path / "p1" / "requirements.yaml"
    p1.parent.mkdir()
    p1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - adaptive != 1.0.0
                - adaptive <2
            """,
        ),
    )

    requirements = parse_requirements(p1, verbose=False)

    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert resolved == {
        "adaptive": {
            None: {
                "conda": Spec(
                    name="adaptive",
                    which="conda",
                    pin="!=1.0.0,<2",
                    identifier="17e5d607",
                ),
                "pip": Spec(
                    name="adaptive",
                    which="pip",
                    pin="!=1.0.0,<2",
                    identifier="17e5d607",
                ),
            },
        },
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_dot_in_package_name(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    p1 = tmp_path / "p1" / "requirements.yaml"
    p1.parent.mkdir()
    p1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - ruamel.yaml
            """,
        ),
    )
    p1 = maybe_as_toml(toml_or_yaml, p1)

    requirements = parse_requirements(p1, verbose=False)
    assert requirements.requirements == {
        "ruamel.yaml": [
            Spec(name="ruamel.yaml", which="conda", identifier="17e5d607"),
            Spec(name="ruamel.yaml", which="pip", identifier="17e5d607"),
        ],
    }



"""unidep conda-lock tests."""
from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest
from ruamel.yaml import YAML

from unidep._conda_lock import (
    LockSpec,
    _handle_missing_keys,
    conda_lock_command,
)
from unidep.utils import remove_top_comments

if TYPE_CHECKING:
    from unidep.platform_definitions import CondaPip, Platform


def test_conda_lock_command(tmp_path: Path) -> None:
    folder = tmp_path / "simple_monorepo"
    shutil.copytree(Path(__file__).parent / "simple_monorepo", folder)
    with patch("unidep._conda_lock._run_conda_lock", return_value=None):
        conda_lock_command(
            depth=1,
            directory=folder,
            platforms=["linux-64", "osx-arm64"],
            verbose=True,
            only_global=False,
            check_input_hash=True,
            ignore_pins=[],
            overwrite_pins=[],
            skip_dependencies=[],
        )
    with YAML(typ="safe") as yaml:
        with (folder / "project1" / "conda-lock.yml").open() as f:
            lock1 = yaml.load(f)
        with (folder / "project2" / "conda-lock.yml").open() as f:
            lock2 = yaml.load(f)

    assert [p["name"] for p in lock1["package"] if p["platform"] == "osx-arm64"] == [
        "bzip2",
        "python_abi",
        "tzdata",
    ]
    assert [p["name"] for p in lock2["package"] if p["platform"] == "osx-arm64"] == [
        "python_abi",
        "tzdata",
    ]


def test_conda_lock_command_pip_package_with_conda_dependency(tmp_path: Path) -> None:
    folder = tmp_path / "test-pip-package-with-conda-dependency"
    shutil.copytree(
        Path(__file__).parent / "test-pip-package-with-conda-dependency",
        folder,
    )
    with patch("unidep._conda_lock._run_conda_lock", return_value=None):
        conda_lock_command(
            depth=1,
            directory=folder,
            platforms=["linux-64"],
            verbose=True,
            only_global=False,
            check_input_hash=True,
            ignore_pins=[],
            overwrite_pins=[],
            skip_dependencies=[],
        )
    with YAML(typ="safe") as yaml:
        with (folder / "project1" / "conda-lock.yml").open() as f:
            lock1 = yaml.load(f)
        with (folder / "project2" / "conda-lock.yml").open() as f:
            lock2 = yaml.load(f)
    assert [p["name"] for p in lock1["package"]] == [
        "_libgcc_mutex",
        "_openmp_mutex",
        "bzip2",
        "ca-certificates",
        "ld_impl_linux-64",
        "libexpat",
        "libffi",
        "libgcc-ng",
        "libgomp",
        "libnsl",
        "libsqlite",
        "libstdcxx-ng",
        "libuuid",
        "libzlib",
        "ncurses",
        "openssl",
        "pybind11",
        "pybind11-global",
        "python",
        "python_abi",
        "readline",
        "tk",
        "tzdata",
        "xz",
    ]
    assert [p["name"] for p in lock2["package"]] == [
        "_libgcc_mutex",
        "_openmp_mutex",
        "bzip2",
        "ca-certificates",
        "ld_impl_linux-64",
        "libexpat",
        "libffi",
        "libgcc-ng",
        "libgomp",
        "libnsl",
        "libsqlite",
        "libstdcxx-ng",
        "libuuid",
        "libzlib",
        "ncurses",
        "openssl",
        "pybind11",
        "pybind11-global",
        "python",
        "python_abi",
        "readline",
        "tk",
        "tzdata",
        "xz",
        "cutde",
        "mako",
        "markupsafe",
        "rsync-time-machine",
    ]


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_conda_lock_command_pip_and_conda_different_name(
    tmp_path: Path,
    capsys: pytest.CaptureFixture,
) -> None:
    folder = tmp_path / "test-pip-and-conda-different-name"
    shutil.copytree(Path(__file__).parent / "test-pip-and-conda-different-name", folder)

    with patch("unidep._conda_lock._run_conda_lock", return_value=None):
        conda_lock_command(
            depth=1,
            directory=folder,
            platforms=["linux-64"],
            verbose=True,
            only_global=False,
            check_input_hash=True,
            ignore_pins=[],
            overwrite_pins=[],
            skip_dependencies=[],
        )
    assert "Missing keys" not in capsys.readouterr().out


def test_remove_top_comments(tmp_path: Path) -> None:
    test_file = tmp_path / "test_file.txt"
    test_file.write_text(
        "# Comment line 1\n# Comment line 2\nActual content line 1\nActual content line 2",
    )

    remove_top_comments(test_file)

    with test_file.open("r") as file:
        content = file.read()

    assert content == "Actual content line 1\nActual content line 2"


def test_handle_missing_keys(capsys: pytest.CaptureFixture) -> None:
    lock_spec = LockSpec(
        packages={
            ("conda", "linux-64", "python-nonexistent"): {
                "name": "python-nonexistent",
                "manager": "conda",
                "platform": "linux-64",
                "dependencies": [],
                "url": "https://example.com/nonexistent",
            },
        },
        dependencies={("conda", "linux-64", "nonexistent"): set()},
    )
    # Here the package name on pip contains the conda package name, so we will download
    # the conda package to verify that this is our package.

    locked: list[dict[str, Any]] = []
    locked_keys: set[tuple[CondaPip, Platform, str]] = {}  # type: ignore[assignment]
    missing_keys: set[tuple[CondaPip, Platform, str]] = {
        ("pip", "linux-64", "nonexistent"),
    }
    with patch(
        "unidep._conda_lock._download_and_get_package_names",
        return_value=None,
    ) as mock:
        _handle_missing_keys(
            lock_spec=lock_spec,
            locked_keys=locked_keys,
            missing_keys=missing_keys,
            locked=locked,
        )
        mock.assert_called_once()

    assert f"❌ Missing keys {missing_keys}" in capsys.readouterr().out
    assert ("pip", "linux-64", "nonexistent") in missing_keys



"""Tests for the version conflict resolution logic."""
from __future__ import annotations

import pytest

from unidep._conflicts import (
    VersionConflictError,
    _combine_pinning_within_platform,
    _is_redundant,
    _is_valid_pinning,
    _parse_pinning,
    combine_version_pinnings,
)
from unidep.platform_definitions import Spec


def test_combining_versions() -> None:
    data = {
        None: {
            "conda": [
                Spec(name="numpy", which="conda", pin=">1"),
                Spec(name="numpy", which="conda", pin="<2"),
            ],
        },
    }
    resolved = _combine_pinning_within_platform(data)  # type: ignore[arg-type]
    assert resolved == {
        None: {
            "conda": Spec(name="numpy", which="conda", pin=">1,<2"),
        },
    }


@pytest.mark.parametrize("operator", ["<", "<=", ">", ">=", "="])
@pytest.mark.parametrize("version", ["1", "1.0", "1.0.0", "1.0.0rc1"])
def test_is_valid_pinning(operator: str, version: str) -> None:
    assert _is_valid_pinning(f"{operator}{version}")


@pytest.mark.parametrize(
    ("pinnings", "expected"),
    [
        ([" > 0.0.1", " < 2", " = 1.0.0"], "=1.0.0"),
        (["<2", ">1"], "<2,>1"),
        ([">1", "<2"], ">1,<2"),
        (["<3", "<=3", "<4"], "<3"),
        (["=1", "=1"], "=1"),
        (["=2", "<3", "<=3", "<4"], "=2"),
        (["=2", ">1", "<3"], "=2"),
        (["=3", ">=2", "<=4"], "=3"),
        (["=3", ">1", "<4"], "=3"),
        (["=3", ">2", "<4"], "=3"),
        ([">=1", "<=1"], ">=1,<=1"),
        ([">=1", ">=1", "=1"], "=1"),
        ([">=1", ">0", "<=3", "<4"], ">=1,<=3"),
        ([">=1", ">0", "<=3", "<4", "!=1.5"], ">=1,<=3,!=1.5"),
        ([">=2", "<=2"], ">=2,<=2"),
        ([">=2", "<3"], ">=2,<3"),
        ([">0.0.1", "<2", "=1.0.0"], "=1.0.0"),
        ([">1", "<=3", "<4"], ">1,<=3"),
        ([">1", "<=3"], ">1,<=3"),
        # TODO #67: !=5 should be removed but this is not yet implemented  # noqa: TD004, FIX002, TD003
        # However, this is not a problem here because !=5 is redundant
        # as it is outside the range of >1 and <=3
        ([">1", "<=3", "!=5"], ">1,<=3,!=5"),
        ([">1", ">=1", "<3", "<=3", ""], ">1,<3"),
        ([">1"], ">1"),
        ([], ""),
    ],
)
def test_combine_version_pinnings(pinnings: list[str], expected: str) -> None:
    assert combine_version_pinnings(pinnings) == expected
    # Try reversing the order of the pinnings
    if "," not in expected:
        assert combine_version_pinnings(pinnings[::-1]) == expected
    else:
        parts = expected.split(",")
        assert combine_version_pinnings(pinnings[::-1]) == ",".join(parts[::-1])


@pytest.mark.parametrize(
    "pinnings",
    [
        ["abc", "def"],
        ["==abc", ">2"],
        ["<=>abc", ">2"],
        [">1", "abc", "<=3", ""],
        ["abc", ">=1", "<=2"],
        ["3", "6"],
        [">", "<"],
    ],
)
def test_invalid_pinnings(pinnings: list[str]) -> None:
    with pytest.raises(VersionConflictError, match="Invalid version pinning"):
        assert combine_version_pinnings(pinnings)


@pytest.mark.parametrize(
    "pinnings",
    [[">2", "<1"], ["<1", ">2"], [">1", "<1"], ["<=1", ">1"], [">1", "<=1"]],
)
def test_contradictory_pinnings(pinnings: list[str]) -> None:
    p1, p2 = pinnings
    with pytest.raises(
        VersionConflictError,
        match=f"Contradictory version pinnings found for `None`: {p1} and {p2}",
    ):
        combine_version_pinnings(pinnings)


def test_exact_pinning_with_contradictory_ranges() -> None:
    with pytest.raises(
        VersionConflictError,
        match="Contradictory version pinnings found for `None`: =3 and <2",
    ):
        combine_version_pinnings(["=3", "<2", ">4"])

    with pytest.raises(
        VersionConflictError,
        match="Contradictory version pinnings found for `None`: =3 and <1",
    ):
        assert combine_version_pinnings(["=3", "<1", ">4"])


def test_multiple_exact_pinnings() -> None:
    with pytest.raises(
        VersionConflictError,
        match="Multiple exact version pinnings found: =2, =3",
    ):
        combine_version_pinnings(["=2", "=3"])


def test_general_contradictory_pinnings() -> None:
    # This test ensures that contradictory non-exact pinnings raise a VersionConflictError
    with pytest.raises(
        VersionConflictError,
        match="Contradictory version pinnings found for `None`: >=2 and <1",
    ):
        combine_version_pinnings([">=2", "<1"])


def test_is_redundant() -> None:
    assert _is_redundant(">2", [">5"])
    assert not _is_redundant(">5", [">2"])
    assert _is_redundant("<5", ["<2"])
    assert _is_redundant(">=2", [">2"])
    assert not _is_redundant(">2", [">=2"])


@pytest.mark.parametrize("pinning", ["<<1", ">>1", "=<1", "=>1"])
def test_invalid_parse_pinning(pinning: str) -> None:
    with pytest.raises(
        VersionConflictError,
        match=f"Invalid version pinning: '{pinning}'",
    ):
        _parse_pinning(pinning)



#!/usr/bin/env python3
"""unidep - Unified Conda and Pip requirements management.

This module provides setuptools integration for unidep.
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from unidep._conflicts import resolve_conflicts
from unidep._dependencies_parsing import parse_requirements
from unidep.utils import (
    UnsupportedPlatformError,
    build_pep508_environment_marker,
    dependencies_filename,
    identify_current_platform,
)

if TYPE_CHECKING:
    import sys

    from setuptools import Distribution

    from unidep.platform_definitions import (
        CondaPip,
        Platform,
        Spec,
    )

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal


def filter_python_dependencies(
    resolved: dict[str, dict[Platform | None, dict[CondaPip, Spec]]],
) -> list[str]:
    """Filter out conda dependencies and return only pip dependencies.

    Examples
    --------
    >>> requirements = parse_requirements("requirements.yaml")
    >>> resolved = resolve_conflicts(
    ...     requirements.requirements, requirements.platforms
    ... )
    >>> python_deps = filter_python_dependencies(resolved)
    """
    pip_deps = []
    for platform_data in resolved.values():
        to_process: dict[Platform | None, Spec] = {}  # platform -> Spec
        for _platform, sources in platform_data.items():
            pip_spec = sources.get("pip")
            if pip_spec:
                to_process[_platform] = pip_spec
        if not to_process:
            continue

        # Check if all Spec objects are identical
        first_spec = next(iter(to_process.values()))
        if all(spec == first_spec for spec in to_process.values()):
            # Build a single combined environment marker
            dep_str = first_spec.name_with_pin(is_pip=True)
            if _platform is not None:
                selector = build_pep508_environment_marker(list(to_process.keys()))  # type: ignore[arg-type]
                dep_str = f"{dep_str}; {selector}"
            pip_deps.append(dep_str)
            continue

        for _platform, pip_spec in to_process.items():
            dep_str = pip_spec.name_with_pin(is_pip=True)
            if _platform is not None:
                selector = build_pep508_environment_marker([_platform])
                dep_str = f"{dep_str}; {selector}"
            pip_deps.append(dep_str)
    return sorted(pip_deps)


def get_python_dependencies(
    filename: str
    | Path
    | Literal["requirements.yaml", "pyproject.toml"] = "requirements.yaml",  # noqa: PYI051
    *,
    verbose: bool = False,
    ignore_pins: list[str] | None = None,
    overwrite_pins: list[str] | None = None,
    skip_dependencies: list[str] | None = None,
    platforms: list[Platform] | None = None,
    raises_if_missing: bool = True,
) -> list[str]:
    """Extract Python (pip) requirements from a `requirements.yaml` or `pyproject.toml` file."""  # noqa: E501
    p = Path(filename)
    if not p.exists():
        if raises_if_missing:
            msg = f"File {filename} not found."
            raise FileNotFoundError(msg)
        return []

    requirements = parse_requirements(
        p,
        ignore_pins=ignore_pins,
        overwrite_pins=overwrite_pins,
        skip_dependencies=skip_dependencies,
        verbose=verbose,
    )
    resolved = resolve_conflicts(
        requirements.requirements,
        platforms or list(requirements.platforms),
    )
    return filter_python_dependencies(resolved)


def _setuptools_finalizer(dist: Distribution) -> None:  # pragma: no cover
    """Entry point called by setuptools to get the dependencies for a project."""
    # PEP 517 says that "All hooks are run with working directory set to the
    # root of the source tree".
    project_root = Path().resolve()
    try:
        requirements_file = dependencies_filename(project_root)
    except FileNotFoundError:
        return
    if requirements_file.exists() and dist.install_requires:
        msg = (
            "You have a `requirements.yaml` file in your project root or"
            " configured unidep in `pyproject.toml` with `[tool.unidep]`,"
            " but you are also using setuptools' `install_requires`."
            " Remove the `install_requires` line from `setup.py`."
        )
        raise RuntimeError(msg)
    try:
        platforms = [identify_current_platform()]
    except UnsupportedPlatformError:
        # We don't know the current platform, so we can't filter out.
        # This will result in selecting all platforms. But this is better
        # than failing.
        platforms = None

    dist.install_requires = get_python_dependencies(
        requirements_file,
        platforms=platforms,
        raises_if_missing=False,
    )



"""unidep CLI tests."""
from __future__ import annotations

import re
import shutil
import subprocess
from pathlib import Path
from unittest.mock import patch

import pytest

from unidep._cli import _install_all_command, _install_command, _pip_compile_command

REPO_ROOT = Path(__file__).parent.parent

EXAMPLE_PROJECTS = [
    "setup_py_project",
    "setuptools_project",
    "hatch_project",
    "pyproject_toml_project",
    "hatch2_project",
]


@pytest.mark.parametrize(
    "project",
    EXAMPLE_PROJECTS,
)
def test_install_command(project: str, capsys: pytest.CaptureFixture) -> None:
    _install_command(
        REPO_ROOT / "example" / project,
        conda_executable="",
        dry_run=True,
        editable=False,
        verbose=True,
    )
    captured = capsys.readouterr()
    assert "Installing conda dependencies" in captured.out
    assert "Installing pip dependencies" in captured.out
    assert "Installing project with" in captured.out


@pytest.mark.parametrize(
    "project",
    EXAMPLE_PROJECTS,
)
def test_unidep_install_dry_run(project: str) -> None:
    # Path to the requirements file
    requirements_path = REPO_ROOT / "example" / project

    # Ensure the requirements file exists
    assert requirements_path.exists(), "Requirements file does not exist"

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607, S603
            "unidep",
            "install",
            "--dry-run",
            str(requirements_path),
        ],
        check=True,
        capture_output=True,
        text=True,
    )

    # Check the output
    assert result.returncode == 0, "Command failed to execute successfully"
    if project in ("setup_py_project", "setuptools_project"):
        assert "📦 Installing conda dependencies with" in result.stdout
    assert "📦 Installing pip dependencies with" in result.stdout
    assert "📦 Installing project with" in result.stdout


def test_install_all_command(capsys: pytest.CaptureFixture) -> None:
    _install_all_command(
        conda_executable="",
        dry_run=True,
        editable=True,
        directory=REPO_ROOT / "example",
        depth=1,
        verbose=False,
    )
    captured = capsys.readouterr()
    assert "Installing conda dependencies" in captured.out
    assert "Installing pip dependencies" in captured.out
    projects = [f"{REPO_ROOT}/example/{p}" for p in EXAMPLE_PROJECTS]
    pkgs = " ".join([f"-e {p}" for p in sorted(projects)])
    assert f"pip install --no-dependencies {pkgs}`" in captured.out


def test_unidep_install_all_dry_run() -> None:
    # Path to the requirements file
    requirements_path = REPO_ROOT / "example"

    # Ensure the requirements file exists
    assert requirements_path.exists(), "Requirements file does not exist"

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607, S603
            "unidep",
            "install-all",
            "--dry-run",
            "--editable",
            "--directory",
            str(requirements_path),
        ],
        check=True,
        capture_output=True,
        text=True,
    )

    # Check the output
    assert result.returncode == 0, "Command failed to execute successfully"
    assert "📦 Installing conda dependencies with `" in result.stdout

    assert r"📦 Installing pip dependencies with `" in result.stdout
    assert (
        "📝 Found local dependencies: {'pyproject_toml_project': ['hatch_project'], 'setup_py_project': ['hatch_project', 'setuptools_project'], 'setuptools_project': ['hatch_project']}"
        in result.stdout
    )
    projects = [f"{REPO_ROOT}/example/{p}" for p in EXAMPLE_PROJECTS]
    pkgs = " ".join([f"-e {p}" for p in sorted(projects)])
    assert "📦 Installing project with `" in result.stdout
    assert f" -m pip install --no-dependencies {pkgs}" in result.stdout


def test_doubly_nested_project_folder_installable(
    tmp_path: Path,
) -> None:
    example_folder = tmp_path / "example"
    shutil.copytree(REPO_ROOT / "example", example_folder)

    # Add an extra project
    extra_projects = example_folder / "extra_projects"
    extra_projects.mkdir(exist_ok=True, parents=True)
    project4 = extra_projects / "project4"
    project4.mkdir(exist_ok=True, parents=True)
    (project4 / "requirements.yaml").write_text(
        "local_dependencies: [../../setup_py_project]",
    )
    pyproject_toml = "\n".join(  # noqa: FLY002
        (
            "[build-system]",
            'requires = ["setuptools", "unidep"]',
            'build-backend = "setuptools.build_meta"',
        ),
    )

    (project4 / "pyproject.toml").write_text(pyproject_toml)
    setup = "\n".join(  # noqa: FLY002
        (
            "from setuptools import setup",
            'setup(name="project4", version="0.1.0", description="yolo", py_modules=["setup_py_project"])',
        ),
    )
    (project4 / "setup.py").write_text(setup)
    (project4 / "project4.py").write_text("print('hello')")

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607, S603
            "unidep",
            "install",
            "--dry-run",
            "--editable",
            "--no-dependencies",
            str(project4 / "requirements.yaml"),
        ],
        check=True,
        capture_output=True,
        text=True,
    )

    p1 = f"{tmp_path}/example/hatch_project"
    p2 = f"{tmp_path}/example/setup_py_project"
    p3 = f"{tmp_path}/example/setuptools_project"
    p4 = f"{tmp_path}/example/extra_projects/project4"
    pkgs = " ".join([f"-e {p}" for p in sorted((p1, p2, p3, p4))])
    assert f"pip install --no-dependencies {pkgs}`" in result.stdout

    p5 = f"{tmp_path}/example/pyproject_toml_project"
    p6 = f"{tmp_path}/example/hatch2_project"
    # Test depth 2
    result = subprocess.run(
        [  # noqa: S607, S603
            "unidep",
            "install-all",
            "--dry-run",
            "--editable",
            "--no-dependencies",
            "--directory",
            example_folder,
            "--depth",
            "2",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    pkgs = " ".join([f"-e {p}" for p in sorted((p1, p2, p3, p4, p5, p6))])
    assert f"pip install --no-dependencies {pkgs}`" in result.stdout

    # Test depth 1 (should not install project4)
    result = subprocess.run(
        [  # noqa: S607, S603
            "unidep",
            "install-all",
            "--dry-run",
            "--editable",
            "--no-dependencies",
            "--directory",
            example_folder,
            "--depth",
            "1",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    pkgs = " ".join([f"-e {p}" for p in sorted((p1, p2, p3, p5, p6))])
    assert f"pip install --no-dependencies {pkgs}`" in result.stdout


def test_pip_compile_command(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    folder = tmp_path / "example"
    shutil.copytree(REPO_ROOT / "example", folder)
    with patch("subprocess.run", return_value=None), patch(
        "importlib.util.find_spec",
        return_value=True,
    ):
        _pip_compile_command(
            depth=2,
            directory=folder,
            platform="linux-64",
            ignore_pins=[],
            skip_dependencies=[],
            overwrite_pins=[],
            verbose=True,
            extra_flags=["--", "--allow-unsafe"],
        )
    requirements_in = folder / "requirements.in"
    assert requirements_in.exists()
    with requirements_in.open() as f:
        assert "adaptive" in f.read()
    requirements_txt = folder / "requirements.txt"

    assert (
        f"Locking dependencies with `pip-compile --output-file {requirements_txt} --allow-unsafe {requirements_in}`"
        in capsys.readouterr().out
    )


def test_install_non_existing_file() -> None:
    with pytest.raises(FileNotFoundError, match="File `does_not_exist` not found."):
        _install_command(
            Path("does_not_exist"),
            conda_executable="",
            dry_run=True,
            editable=True,
            verbose=True,
        )


def test_install_non_existing_folder(tmp_path: Path) -> None:
    match = re.escape(
        f"File `{tmp_path}/requirements.yaml` or `{tmp_path}/pyproject.toml`"
        f" (with unidep configuration) not found in `{tmp_path}`",
    )
    with pytest.raises(FileNotFoundError, match=match):
        _install_command(
            tmp_path,
            conda_executable="",
            dry_run=True,
            editable=True,
            verbose=True,
        )



"""unidep - Unified Conda and Pip requirements management.

This module provides parsing of `requirements.yaml` and `pyproject.toml` files.
"""

from __future__ import annotations

import hashlib
import os
import sys
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap, CommentedSeq

from unidep.platform_definitions import Platform, Spec, platforms_from_selector
from unidep.utils import (
    dependencies_filename,
    is_pip_installable,
    parse_package_str,
    selector_from_comment,
    unidep_configured_in_toml,
    warn,
)

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import Literal
    else:  # pragma: no cover
        from typing_extensions import Literal


try:  # pragma: no cover
    if sys.version_info >= (3, 11):
        import tomllib
    else:
        import tomli as tomllib
    HAS_TOML = True
except ImportError:  # pragma: no cover
    HAS_TOML = False


def find_requirements_files(
    base_dir: str | Path = ".",
    depth: int = 1,
    *,
    verbose: bool = False,
) -> list[Path]:
    """Scan a directory for `requirements.yaml` and `pyproject.toml` files."""
    base_path = Path(base_dir)
    found_files = []

    # Define a helper function to recursively scan directories
    def _scan_dir(path: Path, current_depth: int) -> None:
        if verbose:
            print(f"🔍 Scanning in `{path}` at depth {current_depth}")
        if current_depth > depth:
            return
        for child in path.iterdir():
            if child.is_dir():
                _scan_dir(child, current_depth + 1)
            elif child.name == "requirements.yaml":
                found_files.append(child)
                if verbose:
                    print(f'🔍 Found `"requirements.yaml"` at `{child}`')
            elif child.name == "pyproject.toml" and unidep_configured_in_toml(child):
                if verbose:
                    print(f'🔍 Found `"pyproject.toml"` with dependencies at `{child}`')
                found_files.append(child)

    _scan_dir(base_path, 0)
    return sorted(found_files)


def _extract_first_comment(
    commented_map: CommentedMap,
    index_or_key: int | str,
) -> str | None:
    """Extract the first comment from a CommentedMap."""
    comments = commented_map.ca.items.get(index_or_key, None)
    if comments is None:
        return None
    comment_strings = next(
        c.value.split("\n")[0].rstrip().lstrip() for c in comments if c is not None
    )
    if not comment_strings:
        # empty string
        return None
    return "".join(comment_strings)


def _identifier(identifier: int, selector: str | None) -> str:
    """Return a unique identifier based on the comment."""
    platforms = None if selector is None else tuple(platforms_from_selector(selector))
    data_str = f"{identifier}-{platforms}"
    # Hash using SHA256 and take the first 8 characters for a shorter hash
    return hashlib.sha256(data_str.encode()).hexdigest()[:8]


def _parse_dependency(
    dependency: str,
    dependencies: CommentedMap,
    index_or_key: int | str,
    which: Literal["conda", "pip", "both"],
    identifier: int,
    ignore_pins: list[str],
    overwrite_pins: dict[str, str | None],
    skip_dependencies: list[str],
) -> list[Spec]:
    name, pin, selector = parse_package_str(dependency)
    if name in ignore_pins:
        pin = None
    if name in skip_dependencies:
        return []
    if name in overwrite_pins:
        pin = overwrite_pins[name]
    comment = (
        _extract_first_comment(dependencies, index_or_key)
        if isinstance(dependencies, (CommentedMap, CommentedSeq))
        else None
    )
    if comment and selector is None:
        selector = selector_from_comment(comment)
    identifier_hash = _identifier(identifier, selector)
    if which == "both":
        return [
            Spec(name, "conda", pin, identifier_hash, selector),
            Spec(name, "pip", pin, identifier_hash, selector),
        ]
    return [Spec(name, which, pin, identifier_hash, selector)]


class ParsedRequirements(NamedTuple):
    """Requirements with comments."""

    channels: list[str]
    platforms: list[Platform]
    requirements: dict[str, list[Spec]]


class Requirements(NamedTuple):
    """Requirements as CommentedSeq."""

    # mypy doesn't support CommentedSeq[str], so we use list[str] instead.
    channels: list[str]  # actually a CommentedSeq[str]
    conda: list[str]  # actually a CommentedSeq[str]
    pip: list[str]  # actually a CommentedSeq[str]


def _parse_overwrite_pins(overwrite_pins: list[str]) -> dict[str, str | None]:
    """Parse overwrite pins."""
    result = {}
    for overwrite_pin in overwrite_pins:
        pkg = parse_package_str(overwrite_pin)
        result[pkg.name] = pkg.pin
    return result


def _load(p: Path, yaml: YAML) -> dict[str, Any]:
    if p.suffix == ".toml":
        if not HAS_TOML:  # pragma: no cover
            msg = (
                "❌ No toml support found in your Python installation."
                " If you are using unidep from `pyproject.toml` and this"
                " error occurs during installation, make sure you add"
                '\n\n[build-system]\nrequires = [..., "unidep[toml]"]\n\n'
                " Otherwise, please install it with `pip install tomli`."
            )
            raise ImportError(msg)
        with p.open("rb") as f:
            return tomllib.load(f)["tool"]["unidep"]
    with p.open() as f:
        return yaml.load(f)


def _get_local_dependencies(data: dict[str, Any]) -> list[str]:
    """Get `local_dependencies` from a `requirements.yaml` or `pyproject.toml` file."""
    if "local_dependencies" in data:
        return data["local_dependencies"]
    if "includes" in data:
        warn(
            "⚠️ You are using `includes` in `requirements.yaml` or `pyproject.toml`"
            " `[unidep.tool]` which is deprecated since 0.42.0 and has been renamed to"
            " `local_dependencies`.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return data["includes"]
    return []


def parse_requirements(  # noqa: PLR0912
    *paths: Path,
    ignore_pins: list[str] | None = None,
    overwrite_pins: list[str] | None = None,
    skip_dependencies: list[str] | None = None,
    verbose: bool = False,
) -> ParsedRequirements:
    """Parse a list of `requirements.yaml` or `pyproject.toml` files."""
    ignore_pins = ignore_pins or []
    skip_dependencies = skip_dependencies or []
    overwrite_pins_map = _parse_overwrite_pins(overwrite_pins or [])
    requirements: dict[str, list[Spec]] = defaultdict(list)
    channels: set[str] = set()
    platforms: set[Platform] = set()
    datas = []
    seen: set[Path] = set()
    yaml = YAML(typ="rt")
    for p in paths:
        if verbose:
            print(f"📄 Parsing `{p}`")
        data = _load(p, yaml)
        datas.append(data)
        seen.add(p.resolve())

        # Handle "local_dependencies" (or old name "includes", changed in 0.42.0)
        for include in _get_local_dependencies(data):
            try:
                requirements_path = dependencies_filename(p.parent / include).resolve()
            except FileNotFoundError:
                # Means that this is a local package that is not managed by unidep.
                # We do not need to do anything here, just in `unidep install`.
                continue
            if requirements_path in seen:
                continue  # Avoids circular local_dependencies
            if verbose:
                print(f"📄 Parsing `{include}` from `local_dependencies`")
            datas.append(_load(requirements_path, yaml))
            seen.add(requirements_path)

    identifier = -1
    for data in datas:
        for channel in data.get("channels", []):
            channels.add(channel)
        for _platform in data.get("platforms", []):
            platforms.add(_platform)
        if "dependencies" not in data:
            continue
        dependencies = data["dependencies"]
        for i, dep in enumerate(data["dependencies"]):
            identifier += 1
            if isinstance(dep, str):
                specs = _parse_dependency(
                    dep,
                    dependencies,
                    i,
                    "both",
                    identifier,
                    ignore_pins,
                    overwrite_pins_map,
                    skip_dependencies,
                )
                for spec in specs:
                    requirements[spec.name].append(spec)
                continue
            assert isinstance(dep, dict)
            for which in ["conda", "pip"]:
                if which in dep:
                    specs = _parse_dependency(
                        dep[which],
                        dep,
                        which,
                        which,  # type: ignore[arg-type]
                        identifier,
                        ignore_pins,
                        overwrite_pins_map,
                        skip_dependencies,
                    )
                    for spec in specs:
                        requirements[spec.name].append(spec)

    return ParsedRequirements(sorted(channels), sorted(platforms), dict(requirements))


# Alias for backwards compatibility
parse_yaml_requirements = parse_requirements


def _extract_local_dependencies(
    path: Path,
    base_path: Path,
    processed: set[Path],
    dependencies: dict[str, set[str]],
    *,
    check_pip_installable: bool = True,
    verbose: bool = False,
) -> None:
    if path in processed:
        return
    processed.add(path)
    yaml = YAML(typ="safe")
    data = _load(path, yaml)
    # Handle "local_dependencies" (or old name "includes", changed in 0.42.0)
    for include in _get_local_dependencies(data):
        assert not os.path.isabs(include)  # noqa: PTH117
        abs_include = (path.parent / include).resolve()
        if not abs_include.exists():
            msg = f"File `{include}` not found."
            raise FileNotFoundError(msg)

        try:
            requirements_path = dependencies_filename(abs_include)
        except FileNotFoundError:
            # Means that this is a local package that is not managed by unidep.
            if is_pip_installable(abs_include):
                dependencies[str(base_path)].add(str(abs_include))
                warn(
                    f"⚠️ Installing a local dependency (`{abs_include.name}`) which is"
                    " not managed by unidep, this will skip all of its dependencies,"
                    " i.e., it will call `pip install` with `--no-dependencies`."
                    " To properly manage this dependency, add a `requirements.yaml`"
                    " or `pyproject.toml` file with `[tool.unidep]` in its directory.",
                )
            else:
                msg = (
                    f"`{include}` in `local_dependencies` is not pip installable nor is"
                    " it managed by unidep. Remove it from `local_dependencies`."
                )
                raise RuntimeError(msg) from None
            continue

        project_path = str(requirements_path.parent)
        if project_path == str(base_path):
            continue
        if not check_pip_installable or (
            is_pip_installable(base_path)
            and is_pip_installable(requirements_path.parent)
        ):
            dependencies[str(base_path)].add(project_path)
        if verbose:
            print(f"🔗 Adding `{requirements_path}` from `local_dependencies`")
        _extract_local_dependencies(
            requirements_path,
            base_path,
            processed,
            dependencies,
            check_pip_installable=check_pip_installable,
        )


def parse_local_dependencies(
    *paths: Path,
    check_pip_installable: bool = True,
    verbose: bool = False,
) -> dict[Path, list[Path]]:
    """Extract local project dependencies from a list of `requirements.yaml` or `pyproject.toml` files.

    Works by loading the specified `local_dependencies` list.
    """  # noqa: E501
    dependencies: dict[str, set[str]] = defaultdict(set)

    for p in paths:
        if verbose:
            print(f"🔗 Analyzing dependencies in `{p}`")
        base_path = p.resolve().parent
        _extract_local_dependencies(
            path=p,
            base_path=base_path,
            processed=set(),
            dependencies=dependencies,
            check_pip_installable=check_pip_installable,
            verbose=verbose,
        )

    return {
        Path(k): sorted({Path(v) for v in v_set})
        for k, v_set in sorted(dependencies.items())
    }


def yaml_to_toml(yaml_path: Path) -> str:
    """Converts a `requirements.yaml` file TOML format."""
    try:
        import tomli_w
    except ImportError:  # pragma: no cover
        msg = (
            "❌ `tomli_w` is required to convert YAML to TOML."
            " Install it with `pip install tomli_w`."
        )
        raise ImportError(msg) from None
    yaml = YAML(typ="rt")
    data = _load(yaml_path, yaml)
    data.pop("name", None)
    dependencies = data.get("dependencies", [])
    for i, dep in enumerate(dependencies):
        if isinstance(dep, str):
            comment = _extract_first_comment(dependencies, i)
            if comment is not None:
                selector = selector_from_comment(comment)
                if selector is not None:
                    dependencies[i] = f"{dep}:{selector}"
            continue
        assert isinstance(dep, dict)
        for which in ["conda", "pip"]:
            if which in dep:
                comment = _extract_first_comment(dep, which)
                if comment is not None:
                    selector = selector_from_comment(comment)
                    if selector is not None:
                        dep[which] = f"{dep[which]}:{selector}"

    return tomli_w.dumps({"tool": {"unidep": data}})



"""unidep - Unified Conda and Pip requirements management.

Verion conflict detections and resolution.
"""

from __future__ import annotations

import sys
from collections import defaultdict
from typing import TYPE_CHECKING

from packaging import version

from unidep.platform_definitions import Platform, Spec
from unidep.utils import warn

if sys.version_info >= (3, 8):
    from typing import get_args
else:  # pragma: no cover
    from typing_extensions import get_args


if TYPE_CHECKING:
    from unidep.platform_definitions import CondaPip

VALID_OPERATORS = ["<=", ">=", "<", ">", "=", "!="]
_REPO_URL = "https://github.com/basnijholt/unidep"


def _prepare_specs_for_conflict_resolution(
    requirements: dict[str, list[Spec]],
) -> dict[str, dict[Platform | None, dict[CondaPip, list[Spec]]]]:
    """Prepare and group metadata for conflict resolution.

    This function groups metadata by platform and source for each package.

    :param requirements: Dictionary mapping package names to a list of Spec objects.
    :return: Dictionary mapping package names to grouped metadata.
    """
    prepared_data = {}
    for package, spec_list in requirements.items():
        grouped_specs: dict[Platform | None, dict[CondaPip, list[Spec]]] = defaultdict(
            lambda: defaultdict(list),
        )
        for spec in spec_list:
            _platforms = spec.platforms()
            if _platforms is None:
                _platforms = [None]  # type: ignore[list-item]
            for _platform in _platforms:
                grouped_specs[_platform][spec.which].append(spec)

        # Convert defaultdicts to dicts
        prepared_data[package] = {k: dict(v) for k, v in grouped_specs.items()}
    return prepared_data


def _pop_unused_platforms_and_maybe_expand_none(
    platform_data: dict[Platform | None, dict[CondaPip, list[Spec]]],
    platforms: list[Platform] | None,
) -> None:
    """Expand `None` to all platforms if there is a platform besides None."""
    allowed_platforms = get_args(Platform)
    if platforms:
        allowed_platforms = platforms  # type: ignore[assignment]

    # If there is a platform besides None, expand None to all platforms
    if len(platform_data) > 1 and None in platform_data:
        sources = platform_data.pop(None)
        for _platform in allowed_platforms:
            for which, specs in sources.items():
                platform_data.setdefault(_platform, {}).setdefault(which, []).extend(
                    specs,
                )

    # Remove platforms that are not allowed
    to_pop = platform_data.keys() - allowed_platforms
    to_pop.discard(None)
    for _platform in to_pop:
        platform_data.pop(_platform)


def _maybe_new_spec_with_combined_pinnings(
    specs: list[Spec],
) -> Spec:
    pinned_specs = [m for m in specs if m.pin is not None]
    if len(pinned_specs) > 1:
        first = pinned_specs[0]
        pins = [m.pin for m in pinned_specs]
        pin = combine_version_pinnings(pins, name=first.name)  # type: ignore[arg-type]
        return Spec(
            name=first.name,
            which=first.which,
            pin=pin,
            identifier=first.identifier,  # should I create a new one?
        )

    # Flatten the list
    return specs[0]


def _combine_pinning_within_platform(
    data: dict[Platform | None, dict[CondaPip, list[Spec]]],
) -> dict[Platform | None, dict[CondaPip, Spec]]:
    reduced_data: dict[Platform | None, dict[CondaPip, Spec]] = {}
    for _platform, packages in data.items():
        reduced_data[_platform] = {}
        for which, specs in packages.items():
            spec = _maybe_new_spec_with_combined_pinnings(specs)
            reduced_data[_platform][which] = spec
    return reduced_data


def _resolve_conda_pip_conflicts(sources: dict[CondaPip, Spec]) -> dict[CondaPip, Spec]:
    conda_spec = sources.get("conda")
    pip_spec = sources.get("pip")
    if not conda_spec or not pip_spec:  # If either is missing, there is no conflict
        return sources

    # Compare version pins to resolve conflicts
    if conda_spec.pin and not pip_spec.pin:
        return {"conda": conda_spec}  # Prefer conda if it has a pin
    if pip_spec.pin and not conda_spec.pin:
        return {"pip": pip_spec}  # Prefer pip if it has a pin
    if conda_spec.pin == pip_spec.pin:
        return {"conda": conda_spec, "pip": pip_spec}  # Keep both if pins are identical

    # Handle conflict where both conda and pip have different pins
    warn(
        "Version Pinning Conflict:\n"
        f"Different version specifications for Conda ('{conda_spec.pin}') and Pip"
        f" ('{pip_spec.pin}'). Both versions are retained.",
        stacklevel=2,
    )
    return {"conda": conda_spec, "pip": pip_spec}


class VersionConflictError(ValueError):
    """Raised when a version conflict is detected."""


def resolve_conflicts(
    requirements: dict[str, list[Spec]],
    platforms: list[Platform] | None = None,
) -> dict[str, dict[Platform | None, dict[CondaPip, Spec]]]:
    """Resolve conflicts in a dictionary of requirements.

    Uses the ``ParsedRequirements.requirements`` dict returned by
    `parse_requirements`.
    """
    if platforms and not set(platforms).issubset(get_args(Platform)):
        msg = f"Invalid platform: {platforms}, must contain only {get_args(Platform)}"
        raise VersionConflictError(msg)

    prepared = _prepare_specs_for_conflict_resolution(requirements)
    for data in prepared.values():
        _pop_unused_platforms_and_maybe_expand_none(data, platforms)
    resolved = {
        pkg: _combine_pinning_within_platform(data) for pkg, data in prepared.items()
    }

    for _platforms in resolved.values():
        for _platform, sources in _platforms.items():
            _platforms[_platform] = _resolve_conda_pip_conflicts(sources)
    return resolved


def _parse_pinning(pinning: str) -> tuple[str, version.Version]:
    """Separates the operator and the version number."""
    pinning = pinning.strip()
    for operator in VALID_OPERATORS:
        if pinning.startswith(operator):
            version_part = pinning[len(operator) :].strip()
            if version_part:
                try:
                    return operator, version.parse(version_part)
                except version.InvalidVersion:
                    break
            else:
                break  # Empty version string

    msg = f"Invalid version pinning: '{pinning}', must start with one of {VALID_OPERATORS}"  # noqa: E501
    raise VersionConflictError(msg)


def _is_redundant(pinning: str, other_pinnings: list[str]) -> bool:
    """Determines if a version pinning is redundant given a list of other pinnings."""
    op, version = _parse_pinning(pinning)

    for other in other_pinnings:
        other_op, other_version = _parse_pinning(other)
        if other == pinning:
            continue

        if op == "<" and (
            other_op == "<"
            and version >= other_version
            or other_op == "<="
            and version > other_version
        ):
            return True
        if op == "<=" and other_op in ["<", "<="] and version >= other_version:
            return True
        if op == ">" and (
            other_op == ">"
            and version <= other_version
            or other_op == ">="
            and version < other_version
        ):
            return True
        if op == ">=" and other_op in [">", ">="] and version <= other_version:
            return True

    return False


def _is_valid_pinning(pinning: str) -> bool:
    """Checks if a version pinning string is valid."""
    if any(op in pinning for op in VALID_OPERATORS):
        try:
            # Attempt to parse the version part of the pinning
            _parse_pinning(pinning)
            return True  # noqa: TRY300
        except VersionConflictError:
            # If parsing fails, the pinning is not valid
            return False
    # If the pinning doesn't contain any recognized operator, it's not valid
    return False


def _deduplicate(pinnings: list[str]) -> list[str]:
    """Removes duplicate strings."""
    return list(dict.fromkeys(pinnings))  # preserve order


def _split_pinnings(pinnings: list[str]) -> list[str]:
    """Extracts version pinnings from a list of Spec objects."""
    return [_pin.lstrip().rstrip() for pin in pinnings for _pin in pin.split(",")]


def combine_version_pinnings(pinnings: list[str], *, name: str | None = None) -> str:
    """Combines a list of version pinnings into a single string."""
    pinnings = [p for p in pinnings if p != ""]
    pinnings = _split_pinnings(pinnings)
    pinnings = _deduplicate(pinnings)
    if len(pinnings) == 1:
        return pinnings[0]
    for pin in pinnings:
        if not _is_valid_pinning(pin):
            ops = ", ".join(VALID_OPERATORS)
            url = f"{_REPO_URL}/blob/main/README.md#supported-version-pinnings"
            msg = (
                f"Invalid version pinning '{pin}' for '{name}'. "
                "UniDep supports only the following operators for combining pinnings: "
                f"{ops}. For complex pinnings (like VCS URLs, local paths, or build"
                " strings), ensure all pinnings are identical. Divergent complex"
                f" pinnings cannot be combined. See {url} for more information."
            )

            raise VersionConflictError(msg)

    valid_pinnings = [p.replace(" ", "") for p in pinnings]
    exact_pinnings = [p for p in valid_pinnings if p.startswith("=")]
    if len(exact_pinnings) > 1:
        pinnings_str = ", ".join(exact_pinnings)
        msg = f"Multiple exact version pinnings found: {pinnings_str} for `{name}`"
        raise VersionConflictError(msg)

    err_msg = f"Contradictory version pinnings found for `{name}`"

    if exact_pinnings:
        exact_pin = exact_pinnings[0]
        exact_version = version.parse(exact_pin[1:])
        for other_pin in valid_pinnings:
            if other_pin != exact_pin:
                op, ver = _parse_pinning(other_pin)
                if not (
                    (op == "<" and exact_version < ver)
                    or (op == "<=" and exact_version <= ver)
                    or (op == ">" and exact_version > ver)
                    or (op == ">=" and exact_version >= ver)
                ):
                    msg = f"{err_msg}: {exact_pin} and {other_pin}"
                    raise VersionConflictError(msg)
        return exact_pin

    non_redundant_pinnings = [
        pin for pin in valid_pinnings if not _is_redundant(pin, valid_pinnings)
    ]

    for i, pin in enumerate(non_redundant_pinnings):
        for other_pin in non_redundant_pinnings[i + 1 :]:
            op1, ver1 = _parse_pinning(pin)
            op2, ver2 = _parse_pinning(other_pin)
            msg = f"{err_msg}: {pin} and {other_pin}"
            # Check for direct contradictions like >2 and <1
            if (op1 == ">" and op2 == "<" and ver1 >= ver2) or (
                op1 == "<" and op2 == ">" and ver1 <= ver2
            ):
                raise VersionConflictError(msg)

            # Check for contradictions involving inclusive bounds like >=2 and <1
            if (
                (op1 == ">=" and op2 == "<" and ver1 >= ver2)
                or (op1 == ">" and op2 == "<=" and ver1 >= ver2)
                or (op1 == "<=" and op2 == ">" and ver1 <= ver2)
                or (op1 == ">" and op2 == "<=" and ver1 >= ver2)
            ):
                raise VersionConflictError(msg)

    return ",".join(non_redundant_pinnings)



"""unidep - Unified Conda and Pip requirements management.

Types and definitions for platforms, selectors, and markers.
"""
from __future__ import annotations

import sys
from typing import NamedTuple, cast

if sys.version_info >= (3, 8):
    from typing import Literal, get_args
else:  # pragma: no cover
    from typing_extensions import Literal, get_args


CondaPlatform = Literal["unix", "linux", "osx", "win"]
Platform = Literal[
    "linux-64",
    "linux-aarch64",
    "linux-ppc64le",
    "osx-64",
    "osx-arm64",
    "win-64",
]
Selector = Literal[
    "linux64",
    "aarch64",
    "ppc64le",
    "osx64",
    "arm64",
    "win64",
    "win",
    "unix",
    "linux",
    "osx",
    "macos",
]
# The following are also supported in conda-build but not in UniDep:
# "linux-32" (32-bit x86 on Linux)
# "linux-64" (64-bit x86 on Linux)
# "linux-ppc64" (64-bit PowerPC on Linux)
# "linux-ppc64le" (64-bit Little Endian PowerPC on Linux)
# "linux-s390x" (64-bit IBM z Systems on Linux)
# "linux-armv6l" (32-bit ARMv6 on Linux)
# "linux-armv7l" (32-bit ARMv7 on Linux)
# "win-32" (32-bit x86 Windows)
# "win-arm64" (64-bit ARM on Windows)
CondaPip = Literal["conda", "pip"]

VALID_SELECTORS = get_args(Selector)

PEP508_MARKERS = {
    "linux-64": "sys_platform == 'linux' and platform_machine == 'x86_64'",
    "linux-aarch64": "sys_platform == 'linux' and platform_machine == 'aarch64'",
    "linux-ppc64le": "sys_platform == 'linux' and platform_machine == 'ppc64le'",
    "osx-64": "sys_platform == 'darwin' and platform_machine == 'x86_64'",
    "osx-arm64": "sys_platform == 'darwin' and platform_machine == 'arm64'",
    "win-64": "sys_platform == 'win32' and platform_machine == 'AMD64'",
    ("linux-64", "linux-aarch64", "linux-ppc64le"): "sys_platform == 'linux'",
    ("osx-64", "osx-arm64"): "sys_platform == 'darwin'",
    (
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
        "osx-64",
        "osx-arm64",
    ): "sys_platform == 'linux' or sys_platform == 'darwin'",
}


# The first element of each tuple is the only unique selector
PLATFORM_SELECTOR_MAP: dict[Platform, list[Selector]] = {
    "linux-64": ["linux64", "unix", "linux"],
    "linux-aarch64": ["aarch64", "unix", "linux"],
    "linux-ppc64le": ["ppc64le", "unix", "linux"],
    # "osx64" is a selector unique to conda-build referring to
    # platforms on macOS and the Python architecture is x86-64
    "osx-64": ["osx64", "osx", "macos", "unix"],
    "osx-arm64": ["arm64", "osx", "macos", "unix"],
    "win-64": ["win64", "win"],
}

PLATFORM_SELECTOR_MAP_REVERSE: dict[Selector, set[Platform]] = {}
for _platform, _selectors in PLATFORM_SELECTOR_MAP.items():
    for _selector in _selectors:
        PLATFORM_SELECTOR_MAP_REVERSE.setdefault(_selector, set()).add(_platform)


def validate_selector(selector: Selector) -> None:
    """Check if a selector is valid."""
    valid_selectors = VALID_SELECTORS
    if selector not in VALID_SELECTORS:
        msg = f"Invalid platform selector: `{selector}`, use one of `{valid_selectors}`"
        raise ValueError(msg)


def platforms_from_selector(selector: str) -> list[Platform]:
    """Extract platforms from a selector.

    For example, selector can be ``'linux64 win64'`` or ``'osx'``.
    """
    # we support a very limited set of selectors that adhere to platform only
    # refs:
    # https://docs.conda.io/projects/conda-build/en/latest/resources/define-metadata.html#preprocessing-selectors
    # https://github.com/conda/conda-lock/blob/3d2bf356e2cf3f7284407423f7032189677ba9be/conda_lock/src_parser/selectors.py
    platforms: set[Platform] = set()
    for s in selector.split():
        s = cast(Selector, s)
        platforms |= set(PLATFORM_SELECTOR_MAP_REVERSE[s])
    return sorted(platforms)


class Spec(NamedTuple):
    """A dependency specification."""

    name: str
    which: CondaPip
    pin: str | None = None
    identifier: str | None = None
    # can be of type `Selector` but also space separated string of `Selector`s
    selector: str | None = None

    def platforms(self) -> list[Platform] | None:
        """Return the platforms for this dependency."""
        if self.selector is None:
            return None
        return platforms_from_selector(self.selector)

    def pprint(self) -> str:
        """Pretty print the dependency."""
        result = f"{self.name}"
        if self.pin is not None:
            result += f" {self.pin}"
        if self.selector is not None:
            result += f" # [{self.selector}]"
        return result

    def name_with_pin(self, *, is_pip: bool = False) -> str:
        """Return the name with the pin."""
        result = f"{self.name}"
        if self.pin is not None:
            pin = self.pin
            if (
                is_pip
                and "=" in pin
                and not (">=" in pin or "<=" in pin or "==" in pin)
            ):
                # Replace `=` with `==` for pip
                pin = pin.replace("=", "==")
            result += f" {pin}"
        return result



"""unidep's YAML parsing of the `local_dependencies` list."""
from __future__ import annotations

import shutil
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from ruamel.yaml import YAML

from unidep import (
    find_requirements_files,
    parse_local_dependencies,
    parse_requirements,
    resolve_conflicts,
)
from unidep._dependencies_parsing import yaml_to_toml

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:  # pragma: no cover
        from typing_extensions import Literal


REPO_ROOT = Path(__file__).parent.parent


def maybe_as_toml(toml_or_yaml: Literal["toml", "yaml"], p: Path) -> Path:
    if toml_or_yaml == "toml":
        toml = yaml_to_toml(p)
        p.unlink()
        p = p.with_name("pyproject.toml")
        p.write_text(toml)
    return p


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_circular_local_dependencies(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    project2 = tmp_path / "project2"
    project2.mkdir(exist_ok=True, parents=True)

    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - adaptive-scheduler
            local_dependencies:
                - ../project2
                - ../project2  # duplicate include (shouldn't affect the result)
            """,
        ),
    )
    # Test with old `includes` name
    r2.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - adaptive
            includes:  # `local_dependencies` was called `includes` in ≤0.41.0
                - ../project1
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)
    # Only convert r1 to toml, not r2, because we want to test that
    with pytest.warns(DeprecationWarning, match="is deprecated since 0.42.0"):
        requirements = parse_requirements(r1, r2, verbose=False)
    # Both will be duplicated because of the circular dependency
    # but `resolve_conflicts` will remove the duplicates
    assert len(requirements.requirements["adaptive"]) == 4
    assert len(requirements.requirements["adaptive-scheduler"]) == 2
    resolved = resolve_conflicts(requirements.requirements, requirements.platforms)
    assert len(resolved["adaptive"]) == 1
    assert len(resolved["adaptive"][None]) == 2
    assert len(resolved["adaptive-scheduler"]) == 1
    assert len(resolved["adaptive-scheduler"][None]) == 2


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_local_dependencies(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    project2 = tmp_path / "project2"
    project2.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project2
                - ../project2  # duplicate include (shouldn't affect the result)
            """,
        ),
    )
    r2.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project1
            """,
        ),
    )
    r2 = maybe_as_toml(toml_or_yaml, r2)
    # Only convert r2 to toml, not r1, because we want to test that
    local_dependencies = parse_local_dependencies(
        r1,
        r2,
        verbose=False,
        check_pip_installable=False,
    )
    expected_dependencies = {
        project1.resolve(): [project2.resolve()],
        project2.resolve(): [project1.resolve()],
    }
    assert local_dependencies == expected_dependencies


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nested_local_dependencies(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project2 = tmp_path / "project2"
    project3 = tmp_path / "project3"
    project4 = tmp_path / "project4"
    for project in [project1, project2, project3, project4]:
        project.mkdir(exist_ok=True, parents=True)

    p1 = project1 / "requirements.yaml"
    p2 = project2 / "requirements.yaml"
    p3 = project3 / "requirements.yaml"
    p4 = project4 / "requirements.yaml"
    p1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project2
            """,
        ),
    )
    p2.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project3
            """,
        ),
    )
    p3.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project4
            """,
        ),
    )
    p4.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - numpy
            """,
        ),
    )
    p1 = maybe_as_toml(toml_or_yaml, p1)
    p2 = maybe_as_toml(toml_or_yaml, p2)
    p3 = maybe_as_toml(toml_or_yaml, p3)
    p4 = maybe_as_toml(toml_or_yaml, p4)
    local_dependencies = parse_local_dependencies(
        p1,
        p2,
        p3,
        verbose=False,
        check_pip_installable=False,
    )
    expected_dependencies = {
        project1.resolve(): [
            project2.resolve(),
            project3.resolve(),
            project4.resolve(),
        ],
        project2.resolve(): [project3.resolve(), project4.resolve()],
        project3.resolve(): [project4.resolve()],
    }
    assert local_dependencies == expected_dependencies


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nonexistent_local_dependencies(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../nonexistent_project
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)
    with pytest.raises(FileNotFoundError, match="not found."):
        parse_local_dependencies(r1, verbose=False, check_pip_installable=False)


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_no_local_dependencies(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pandas
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)
    local_dependencies = parse_local_dependencies(
        r1,
        verbose=False,
        check_pip_installable=False,
    )
    assert local_dependencies == {}


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_mixed_real_and_placeholder_dependencies(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - scipy
            local_dependencies:
                - ../project1  # Self include (circular dependency)
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)
    local_dependencies = parse_local_dependencies(
        r1,
        verbose=False,
        check_pip_installable=False,
    )
    assert local_dependencies == {}


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_local_dependencies_pip_installable(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    example_folder = tmp_path / "example"
    shutil.copytree(REPO_ROOT / "example", example_folder)

    # Add an extra project
    extra_project = example_folder / "project69"
    extra_project.mkdir(exist_ok=True, parents=True)
    (extra_project / "requirements.yaml").write_text(
        "local_dependencies: [../setup_py_project]",
    )

    # Add a line to project1 local_dependencies
    setup_py_project_req = example_folder / "setup_py_project" / "requirements.yaml"
    yaml = YAML(typ="safe")
    with setup_py_project_req.open("r") as f:
        requirements = yaml.load(f)
    requirements["local_dependencies"].append("../project69")
    with setup_py_project_req.open("w") as f:
        yaml.dump(requirements, f)

    setup_py_project_req = maybe_as_toml(toml_or_yaml, setup_py_project_req)
    found_files = find_requirements_files(example_folder)
    assert len(found_files) == 6

    # Add a common requirements file
    common_requirements = example_folder / "common-requirements.yaml"
    common_requirements.write_text("local_dependencies: [./setup_py_project]")
    common_requirements = maybe_as_toml(toml_or_yaml, common_requirements)
    found_files.append(common_requirements)

    local_dependencies = parse_local_dependencies(
        *found_files,
        check_pip_installable=True,
        verbose=True,
    )
    assert local_dependencies
    assert local_dependencies == {
        example_folder / "setup_py_project": [
            example_folder / "hatch_project",
            example_folder / "setuptools_project",
        ],
        example_folder / "setuptools_project": [
            example_folder / "hatch_project",
        ],
        example_folder / "pyproject_toml_project": [
            example_folder / "hatch_project",
        ],
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_parse_local_dependencies_pip_installable_with_non_installable_project(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    example_folder = tmp_path / "example"
    shutil.copytree(REPO_ROOT / "example", example_folder)

    # Add an extra project
    extra_project = example_folder / "project4"
    extra_project.mkdir(exist_ok=True, parents=True)
    r_extra = extra_project / "requirements.yaml"
    r_extra.write_text("local_dependencies: [../setup_py_project]")
    r_extra = maybe_as_toml(toml_or_yaml, r_extra)

    # Add a line to hatch_project local_dependencies which should
    # make hatch_project depend on setup_py_project, via project4! However, project4 is
    # not `pip installable` so we're testing that path.
    setup_py_project_req = example_folder / "hatch_project" / "requirements.yaml"
    yaml = YAML(typ="safe")
    with setup_py_project_req.open("r") as f:
        requirements = yaml.load(f)
    requirements["local_dependencies"] = ["../project4"]
    with setup_py_project_req.open("w") as f:
        yaml.dump(requirements, f)

    found_files = find_requirements_files(example_folder)
    assert len(found_files) == 6

    local_dependencies = parse_local_dependencies(
        *found_files,
        check_pip_installable=True,
        verbose=True,
    )
    assert local_dependencies
    assert local_dependencies == {
        example_folder / "hatch_project": [
            example_folder / "setup_py_project",
            example_folder / "setuptools_project",
        ],
        example_folder / "setup_py_project": [
            example_folder / "hatch_project",
            example_folder / "setuptools_project",
        ],
        example_folder / "pyproject_toml_project": [
            example_folder / "hatch_project",
            example_folder / "setup_py_project",
            example_folder / "setuptools_project",
        ],
        example_folder / "setuptools_project": [
            example_folder / "hatch_project",
            example_folder / "setup_py_project",
        ],
    }


def test_local_non_unidep_managed_dependency(tmp_path: Path) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    project2 = tmp_path / "project2"
    project2.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project2  # is not managed by unidep
            """,
        ),
    )
    r2 = project2 / "setup.py"  # not managed by unidep
    r2.touch()

    requirements = parse_requirements(r1, verbose=True)  # This should not raise
    assert requirements.requirements == {}

    with pytest.warns(UserWarning, match="not managed by unidep"):
        data = parse_local_dependencies(r1, verbose=True)
    assert data == {project1.resolve(): [project2.resolve()]}


def test_local_non_unidep_and_non_installable_managed_dependency(
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    project2 = tmp_path / "project2"
    project2.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project2  # is not managed by unidep and not installable
            """,
        ),
    )
    with pytest.raises(RuntimeError, match="is not pip installable"):
        parse_local_dependencies(r1, verbose=True)
